#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "book.h"
#include "pngio.h"
#include "enviro.h"

#define MAGIC "NOBOOK1"


typedef struct
{
  char magic[8];
  int npages;
  unsigned int d_bg;
} bookhdr_t;


static int imax2(int a, int b) { return a>b?a:b; }

static int incom(const void *_a, const void *_b)
{
  const int *a= _a, *b= _b;
  return *a-*b;
}

static int has_sfx(char *str, char *sfx)
{
  int Lf, Ls;
  Lf= strlen(sfx);
  Ls= strlen(str);
  if (Ls<Lf) return 0;
  str+= Ls- Lf;
  return !strcmp(str, sfx);
}


static int readfo(int fd, void *buf, int len)
{
  while(len)
  {
    int R;
    R= read(fd, buf, len);
    if (R==0) return 1;
    if (R<0) { if (errno==EINTR) continue; else return 1; }
    buf+= R;
    len-= R;
  }
  return 0;
}

static int writefo(int fd, void *buf, int len)
{
  while(len)
  {
    int R;
    R= write(fd, buf, len);
    if (R==0) return 1;
    if (R<0) { if (errno==EINTR) continue; else return 1; }
    buf+= R;
    len-= R;
  }
  return 0;
}

static book_t* allocbook(book_t **RB)
{
  book_t *B;
  B= calloc(1,sizeof(B[0]));
  if (RB) *RB= B;
  return B;
}

static int mkbookpath(char *name,book_t *B)
{
  char fn[2000];
  if (!storage_dir) return 1;
  sprintf(fn,"%s/%s", storage_dir, name);
  B->sodir= strdup(fn);
  sprintf(fn,"%s/book", B->sodir);
  B->sofil= strdup(fn);
  sprintf(fn,"%s.tmp", B->sofil);
  B->sotmp= strdup(fn);
  return 0;
}

static int allocpages(book_t *B)
{ 
  int sz;
  sz= 2*B->npages*sizeof(B->pages[0]);
  B->pages= malloc(sz);
  return sz;
}

static void hdr_to_book(bookhdr_t *H, book_t *B)
{
  B->npages= H->npages;
  B->d_bg= H->d_bg;
}

int book_read(char *fn, book_t **RB)
{
  int i;
  int sz;
  page_t *P;
  bookhdr_t HDR;
  book_t *B;
  int maxfil, maxtim;

  B= allocbook(RB);
  if (mkbookpath(fn, B)) return 1;

  B->fd= open(B->sofil, O_RDWR);
  if (B->fd<0) return 1;

  if (readfo(B->fd, &HDR, sizeof(HDR))) return 1;
  if (memcmp(HDR.magic, MAGIC, 8)) return 1;
  hdr_to_book(&HDR, B);
  sz= allocpages(B);
  if (readfo(B->fd,B->pages, sz)) return 1;

  maxfil= 0; maxtim= 0;
  for(i=0;i<B->npages*2;i++)
  {
    P= B->pages+i;
    if (!P->utim1) continue;
    maxtim= imax2(P->utim1, maxtim);
    maxfil= imax2(imax2(P->img, P->thu), maxfil);
  }
  B->nexfil= maxfil+1;
  B->nextim= maxtim+1;
  return 0;
}

static int book_write_file(book_t *B,char *fn)
{
  int sz;
  bookhdr_t HDR;

  B->fd= open(fn, O_CREAT | O_TRUNC | O_RDWR, 0600);
  if (B->fd<0) return 1;

  HDR.d_bg= B->d_bg;
  HDR.npages= B->npages;
  strcpy(HDR.magic, MAGIC);
  if (writefo(B->fd, &HDR, sizeof(HDR))) return 1;
  sz= 2*B->npages*sizeof(page_t);
  if (writefo(B->fd, B->pages, sz)) return 1;

  return 0;
}

int book_create(char *fn, book_t **RB)
{
  book_t *B;
  int sz;

  B= allocbook(RB);
  if (mkbookpath(fn, B)) return 1;
  if (mkdir(B->sodir,0700)) return 1;

  B->d_bg= 0;
  B->npages= 1;
  sz= allocpages(B);
  memset(B->pages, 0, sz);
  B->nexfil= 1;
  B->nextim= 1;

  return book_write_file(B, B->sofil);
}

static int entry_valid(book_t *B, int e,int *Rtim)
{
  if (Rtim) *Rtim= B->pages[e].utim1;
  if (B->pages[e].utim1==0) return 0;
  if (B->pages[e].utim1!=B->pages[e].utim2) return 0;
  return 1;
}

static void book_select_entry(book_t *B, int pno, int *Rcur, int *Rold)
{
  int cur, old;
  int a, b;
  int ta, tb;
  a= 2*pno;
  b= 2*pno+1;

  switch( entry_valid(B,a,&ta)*2 + entry_valid(B,b,&tb) )
  {
  case 0: cur= -1; old= a; break;
  case 1: cur=  b; old= a; break;
  case 2: cur=  a; old= b; break;
  case 3: if (ta>tb) { cur= a; old= b; }
               else  { cur= b; old= a; }
          break;
  }
  if (Rcur) *Rcur= cur;
  if (Rold) *Rold= old;
}

static FILE *bookopenfile(book_t *B, int fino, const char *mode)
{
  char fn[2000];
  sprintf(fn,"%s/%d.png", B->sodir, fino);
  return fopen(fn,mode);
}

int book_read_image
  (book_t *B, int pno, int iT,
   unsigned int *Rw, unsigned int *Rh, unsigned int **Rd,
   unsigned int *Rbg)
{
  char* e;
  int s;
  FILE *f;
  page_t *P;
  int fino;

  if (pno<0 || pno>=B->npages) return 1;
  book_select_entry(B, pno, &s, NULL);
  if (s<0) return 1; // no valid entry
  P= B->pages + s;
  if (Rbg)
  {
    if (iT) *Rbg= P->t_bg; 
       else *Rbg= P->i_bg;
  }
  if (Rd)
  {
    if (iT) fino= P->thu;
       else fino= P->img;
    if (!fino) return 1;
    f= bookopenfile(B, fino, "rb");
    if (!f) return 1;
    e= PNGrd_argb_special(f,Rw, Rh, Rd);
    fclose(f);
    if (e) return 1;
  }
  return 0;
}

static int book_update_entry(book_t *B, int e)
{
  off_t pos;
  if (B->fd<0) return 1;
  pos= e*sizeof(page_t) + sizeof(bookhdr_t);
  if (lseek(B->fd, pos,SEEK_SET)==(off_t)-1)
    return 1;
  return writefo(B->fd, B->pages+e, sizeof(B->pages[0]));
}

int book_update_image
  (book_t *B, int pno, int iT,
   unsigned int w, unsigned int h, unsigned int *d, unsigned int bg)
{
  char* e;
  int s, o;
  FILE *f;
  int fino;
  if (pno<0 || pno>=B->npages || B->readonly) return 1;
  book_select_entry(B, pno, &o, &s);

  if (d)
  {
    fino= B->nexfil++;
    f= bookopenfile(B, fino, "wb");
    if (!f) return 1;
    e= PNGwd_argb_special(f,w,h,d);
    fclose(f);
    if (e) return 1;
  }
 
  if (o!=-1)
    B->pages[s]= B->pages[o];
  else
    memset(B->pages+s, 0, sizeof(B->pages[0]));

  B->pages[s].utim1= B->nextim++;
  if (iT)
  {
    if (d)
      B->pages[s].thu= fino;
    B->pages[s].t_bg= bg;
  }
  else
  {
    if (d)
      B->pages[s].img= fino;
    B->pages[s].i_bg= bg;
  }
  B->pages[s].utim2= B->pages[s].utim1; 
  
  return book_update_entry(B, s);
}

static book_t* book_copy(book_t *B,int npages)
{
  book_t *C;
  C= allocbook(NULL);
  *C= *B;
  C->npages= npages;
  allocpages(C);
  return C;
}

static void copy_page_range
   (page_t *dst, int dpos, page_t *src, int spos, int len)
{
  int i;
  for(i=0;i<len;i++)
  {
    dst[(dpos+i)*2+0]= src[(spos+i)*2+0];
    dst[(dpos+i)*2+1]= src[(spos+i)*2+1];
  }
}

static int replace_book_file(book_t *ob, book_t *nb)
{
  int R;
  if (book_write_file(nb, nb->sotmp) || rename(nb->sotmp,ob->sofil))
  {
    close(nb->fd);
    free(nb->pages);
    R= 1;
  }
  else
  {
    close(ob->fd);
    ob->fd= nb->fd;
    free(ob->pages);
    ob->npages= nb->npages;
    ob->pages= nb->pages;
    R= 0;
  }
  free(nb);
  return R;
}

int book_make_page(book_t *B, int pos)
{
  book_t *C;
  if (B->readonly) return 1;
  if (pos<0 || pos>B->npages)
   pos= B->npages;
  C= book_copy(B, B->npages+1);
  copy_page_range(C->pages, 0, B->pages, 0, pos);
  copy_page_range(C->pages, pos+1, B->pages, pos, B->npages-pos);
  memset(C->pages+2*pos, 0, 2*sizeof(C->pages[0]));
  return replace_book_file(B, C);
}

int book_del_page(book_t *B, int pos)
{
  book_t *C;
  if (B->readonly || pos<0 || pos>=B->npages || B->npages==1)
    return 1;
  C= book_copy(B, B->npages-1);
  copy_page_range(C->pages, 0, B->pages, 0, pos);
  copy_page_range(C->pages, pos, B->pages, pos+1, B->npages-pos-1);
  return replace_book_file(B, C);
}

void book_cleanup(book_t *B)
{
  int *files; int nfiles;
  int *expu; int nexpu, sexpu;
  int nent, i, k;
  char fn[2000];
  struct dirent *de;
  DIR *dir;

  if (B->readonly) return ;

  nent= B->npages * 2;                    // each page has two entries
  files= malloc(sizeof(files[0])*nent*2); // each entry has an image and
                                          // a thumbnail
  expu= NULL;
  nfiles= 0;
  for(i=0;i<nent;i++)
   if (entry_valid(B, i,NULL))
   { if (B->pages[i].img) files[nfiles++]= B->pages[i].img;
     if (B->pages[i].thu) files[nfiles++]= B->pages[i].thu; }

  if (!nfiles) goto finish;
  qsort(files, nfiles, sizeof(int), incom);

  sexpu= nfiles * 2; nexpu= 0;
 
  dir= opendir(B->sodir);
  if (!dir) goto finish;
  expu= malloc(sizeof(expu[0])*sexpu);

  while(nexpu<sexpu && (de=readdir(dir)))
  {
    if (!has_sfx(de->d_name,".png")) continue;
    k= atoi(de->d_name);
    if (!bsearch(&k, files, nfiles, sizeof(int), incom))
    {
     expu[nexpu++]= k;
     // printf("will remove %d.png\n", k);
    }
  }
  closedir(dir);

  for(i=0;i<nexpu;i++)
  {
    sprintf(fn,"%s/%d.png", B->sodir, expu[i]);
    unlink(fn);
  }

finish:
  free(files);
  if (expu) free(expu);
}

static void freennull(void *A) { if (A) free(A); }

void book_close(book_t *B)
{
  if (B->fd>=0) close(B->fd);
  freennull(B->pages);
  freennull(B->sodir);
  freennull(B->sofil);
  freennull(B->sotmp);
  memset(B, 0, sizeof(*B));
  B->fd= -1;
  free(B);
}

int book_copy_page(book_t *B, int src, int dst)
{
  int Sc, Do;
  page_t *D, *S;

  if (B->readonly) return 1;
  if (src==dst || src<0 || dst<0 || src>=B->npages || dst>=B->npages)
     return 1;
  book_select_entry(B, src, &Sc, NULL);
  if (Sc==-1) return 1;
  book_select_entry(B, dst, NULL, &Do);

  S= B->pages+ Sc;
  D= B->pages+ Do;

  *D= *S;
  D->utim1= B->nextim++;
  D->utim2= D->utim1;

  return book_update_entry(B, Do);
}

void book_delete(char *bn)
{
  char dn[2000];
  char fn[2000];
  struct dirent *de;
  char *ff;
  DIR *dir;

  if (!storage_dir) return ;
  sprintf(dn,"%s/%s", storage_dir, bn);

  dir= opendir(dn);
  if (!dir) return ;
  while((de=readdir(dir)))
  {
    ff= de->d_name;
    if (!strcmp(ff, ".") || !strcmp(ff,"..")) continue;
    sprintf(fn, "%s/%s", dn, ff);
    unlink(fn);
  }
  closedir(dir);
  rmdir(dn);
}

int book_set_bg(book_t *B, int pno, unsigned int bg)
{
  return book_update_image(B, pno, 0,  1, 1, NULL, bg);
}

int book_set_bgall(book_t *B, unsigned int bg)
{
  int R,i;
  R= 0;
  for(i=0;i<B->npages;i++)
    R|= book_set_bg(B, i, bg);
  return R;
}

int book_permute(book_t *B, int n, int *P)
{
  int i,j;
  book_t *C;
  if (!B || B->readonly || B->npages==1 || n!=B->npages)
    return 1;
  C= book_copy(B, B->npages);
  for(i=0;i<n;i++)
  {
    j= P[i];
    C->pages[2*i+0]= B->pages[2*j+0];
    C->pages[2*i+1]= B->pages[2*j+1];
  }
  return replace_book_file(B, C);
}


