
/**************************************************************************
 Each notebook will be stored in its own directory. Within the directory,
 there will be a file called 'book'. This will contain information about
 the whole notebook.

 Each page will be stored in a separate PNG file. Each thumbnail will also
 be stored in a separate PNG file.

 List of pages will be stored in the 'book' file.

 After saving, the directory will be cleaned up by removing old files.
 
 For each page the following information will be stored:
  - width and height
  - background color / pattern
  - image data

 For each thumbnail, the same will be stored, with one difference. In page
 image data, the background color/pattern is not present. Only the foreground
 pixels are stored. Upon display, the background will be added to the
 foreground image. For thumbnail images, the background is present in the
 image data, for fast display. If the background of the thumbnail and the
 page is different, this signals the system that the thumbnail should be
 recomputed at a convenient time.

 So, here is the page struct:
 **************************************************************************/

typedef struct
{
  unsigned int utim1;
  unsigned int img, thu;    // number representation for the image data.
                   // if img==5, for example, the corresponding
                   // file is 5.png.
  unsigned int i_bg, t_bg;  // background information for image and thumbnail.
                   // if the upper 8 bits are zero, this is a color.
                   // otherwise, it's a pattern number.
  unsigned int i_w, i_h, t_w, t_h;
  unsigned int utim2;
} page_t;


typedef struct
{
  int fd;
  char *sodir;
  char *sofil;
  char *sotmp;
  unsigned int d_bg;  // default background
  int nexfil;         // next file name available
  int nextim;         // next 'utim' value available
  int npages;
  page_t *pages;

  int readonly;
} book_t;

/************************************************************************
 Each book file consists of the following:
  - a header, as defined above
  - an array of page_t objects, two for each page
 When a page is saved, its filename is stored in a page_t object.
 This object is then put into one of the two entries for the page.
 The entry with an invalid utim (0) or the one with the older utim
 is selected for this purpose. If the program gets killed while
 doing this write, the other entry will remain valid. i.e. here is the
 pseudocode for page update:

   k= write_image
   if pagelist[2*pno].utim1<pagelist[2*pno+1].utim1
     sel= 2*pno
   else
     sel= 2*pno+1
   pagelist[sel].utim1= pagelist[sel].utim2= book.nextim++
   write_entry(pagelist[sel])

 When the book is read for the first time, the maximum file number
 and the maximum utim will be computed by looking at the array.
 ************************************************************************/


int book_read(char *fn, book_t **RB);
int book_create(char *fn, book_t **RB);

int book_read_image
  (book_t *B, int pno, int iT,
   unsigned int *Rw, unsigned int *Rh, unsigned int **Rd,
   unsigned int *Rbg);

int book_update_image
  (book_t *B, int pno, int iT,
   unsigned int w, unsigned int h, unsigned int *d, unsigned int bg);

int book_make_page(book_t *B, int pos);
int book_del_page(book_t *B, int pos);

void book_cleanup(book_t *B);

int book_copy_page(book_t *B, int src, int dst);

void book_delete(char *bn);

int book_set_bg(book_t *B, int pno, unsigned int bg);
int book_set_bgall(book_t *B, unsigned int bg);

int book_permute(book_t *B, int n, int *P);

