#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <stdint.h>
#include "gfx.h"
#include "shape.h"
#include "canvas.h"
#include "resize_image.h"
#include "config.h"
#include "dbg.h"
#include "genpat.h"

#define FINGER_THRESHOLD 0

#define TOPBITS 0xff000000

static int init_poly;
static int ca_x, ca_y;
                           // sata is the SAVED Data. It doesn't include
                           // the background.
static uint32_t *ca_data, *ca_sata;
static unsigned int ca_w, ca_h;
static int ca_sel;
static int ca_pen_width[4];
static int colors[4];
static shape_t *ca_shapes[4];
static int ca_cursor=-1, ca_cursor_cx, ca_cursor_cy;
static unsigned int *oplog;
static unsigned int oplogl, oplogs;
static int ca_uxl, ca_uyl;  // CAnvas Update X Least, Y least
static int ca_uxg, ca_uyg;  // CAnvas Update X Greates, Y Greatest

static int er_x, er_y, er_w= 9, er_h= 20;
static unsigned int er_bg= 0x00, er_fg= 0xffFFff;

static int sel_mix, sel_miy, sel_max, sel_may;
static int sel_w, sel_h;
static unsigned int *sel_d;

static int ca_mode;

          /// modified flag, should be renamed..
static int ca_mod;

static unsigned int dumybg= 0;
static unsigned int bp_w=1, bp_h=1, *bp_d= &dumybg, bp_mark= 0xFfffFF;
static unsigned int bp_id;


static int canvas_selmode_event(gfx_event_t *event);

static int imin2(int a,int b) { return a<b? a:b; }
static int imax2(int a,int b) { return a>b? a:b; }
static void osd_fadeout();
static void osd_stop();

typedef struct 
{
  int w,h;
  unsigned int *data;
  unsigned int bg;
} clip_t;

#define NCLIPS 9

static clip_t clipboard[NCLIPS];
static int paste_sprite, paste_w, paste_h, paste_free_data;
static unsigned int *paste_d;

static void save_selected_data()
{
  int i;
  if (clipboard[NCLIPS-1].data) free(clipboard[NCLIPS-1].data);
  for(i=NCLIPS-1;i>0;i--) clipboard[i]= clipboard[i-1];
  clipboard[0].w= sel_w;
  clipboard[0].h= sel_h;
  clipboard[0].data= sel_d;
  clipboard[0].bg= bp_id;
  sel_d= NULL;
}

static void penshape(shape_t **Rs, cfg_pen_t *P)
{
  if (*Rs) free(*Rs);
  *Rs= pen_new(P->width);
}

void canvas_reload_config()
{
  penshape(ca_shapes+0, &config.ptitle);
  penshape(ca_shapes+1, &config.ppen);
  penshape(ca_shapes+2, &config.philight);
  colors[0]= config.ptitle.color   | 0xFF000000;
  colors[1]= config.ppen.color     | 0xFF000000;
  colors[2]= config.philight.color | 0xFF000000;

  ca_pen_width[0]= config.ptitle.width;
  ca_pen_width[1]= config.ppen.width;
  ca_pen_width[2]= config.philight.width;
}


void canvas_init(int *Rw,int *Rh)
{
  gfx_canvas_get_buffer(Wcanvas, &ca_w, &ca_h, &ca_data);
  canvas_reload_config();
  ca_sel= 1;
  *Rw= ca_w; *Rh= ca_h;
  ca_sata= calloc(ca_w*ca_h,sizeof(ca_sata[0]));
}

static void pixel(int x,int y,unsigned int value, unsigned int color)
{
  int S;
  unsigned int k;
  unsigned int P;
  unsigned int pR,pG,pB,pA, cR, cG, cB;
  if (x<0 || x>=ca_w) return ;
  if (y<0 || y>=ca_h) return ;
  k= x+y*ca_w;

  if (value==255)
  {
     ca_data[k]= color;
     ca_sata[k]= color;
  }
  else if (value!=0)
  {
     P= ca_data[k];
     pR= P>>16;     pG= P>>8;     pB= P;
     pR &= 0xff;    pG &= 0xff;   pB &= 0xff;
     cR= color>>16; cG= color>>8; cB= color;
     cR &= 0xff;    cG &= 0xff;   cB &= 0xff;

     pR= (cR*value+pR*(255-value)) / 255;
     pG= (cG*value+pG*(255-value)) / 255;
     pB= (cB*value+pB*(255-value)) / 255;

     ca_data[k]= (pR<<16) | (pG<<8) | (pB) | (color&TOPBITS);

     P= ca_sata[k];
     pA= P >> 24;
     pR= P>>16;     pG= P>>8;     pB= P;
     pR &= 0xff;    pG &= 0xff;   pB &= 0xff;

     pR= (cR*value+pR*pA) / (value+pA);
     pG= (cG*value+pG*pA) / (value+pA);
     pB= (cB*value+pB*pA) / (value+pA);
     pA= pA + value;
     if (pA>255) pA= 255;

     ca_sata[k]= (pR<<16) | (pG<<8) | (pB) | (pA<<24);
  } 

  ca_uxl= imin2(ca_uxl, x);  ca_uxg= imax2(ca_uxg, x);
  ca_uyl= imin2(ca_uyl, y);  ca_uyg= imax2(ca_uyg, y);
}

#include "brushline.c"

static void vline_fg(int x0, int y0, int y1, int color)
{
  int y;
  if (x0<0 || x0>=ca_w) return ;
  y0= imax2(0,y0);
  y1= imin2(ca_h-1,y1);
  ca_uxl= imin2(ca_uxl, x0);  ca_uxg= imax2(ca_uxg, x0);
  ca_uyl= imin2(ca_uyl, y0);  ca_uyg= imax2(ca_uyg, y1);
  for(y=y0;y<=y1;y++) ca_data[x0+y*ca_w]= color;
}

static void vline_bg(int x0, int y0, int y1)
{
  int y;
  if (x0<0 || x0>=ca_w) return ;
  y0= imax2(0,y0);
  y1= imin2(ca_h-1,y1);
  ca_uxl= imin2(ca_uxl, x0);  ca_uxg= imax2(ca_uxg, x0);
  ca_uyl= imin2(ca_uyl, y0);  ca_uyg= imax2(ca_uyg, y1);
  for(y=y0;y<=y1;y++) 
  {
   ca_data[x0+y*ca_w]= bp_d[ x0%bp_w + (y%bp_h)*bp_w];
   ca_sata[x0+y*ca_w]= 0;
  }
}

static void hline_fg(int y0, int x0, int x1, int color)
{
  int x;
  if (y0<0 || y0>=ca_h) return ;
  x0= imax2(0, x0);
  x1= imin2(ca_w-1, x1);
  ca_uxl= imin2(ca_uxl, x0);  ca_uxg= imax2(ca_uxg, x1);
  ca_uyl= imin2(ca_uyl, y0);  ca_uyg= imax2(ca_uyg, y0);
  for(x=x0;x<=x1;x++) ca_data[x+y0*ca_w]= color;
}

static void hline_bg(int y0, int x0, int x1)
{
  int x;
  if (y0<0 || y0>=ca_h) return ;
  x0= imax2(0, x0);
  x1= imin2(ca_w-1, x1);
  ca_uxl= imin2(ca_uxl, x0);  ca_uxg= imax2(ca_uxg, x1);
  ca_uyl= imin2(ca_uyl, y0);  ca_uyg= imax2(ca_uyg, y0);
  for(x=x0;x<=x1;x++) 
  {
    ca_data[x+y0*ca_w]= bp_d[ x%bp_w + (y0%bp_h)*bp_w ];
    ca_sata[x+y0*ca_w]= 0;
  }
}

void eraser_point(int x,int y,int mx, int my)
{
  if (mx==0) goto do_y;
  vline_fg(er_x+mx*er_w, er_y-er_h+my, er_y+er_h+my, er_fg);
  vline_bg(er_x-mx*er_w, er_y-er_h, er_y+er_h);
do_y:
  if (my==0) goto done;
  hline_fg(er_y+my*er_h, er_x-er_w+mx, er_x+er_w+mx, er_fg);
  hline_bg(er_y-my*er_h, er_x-er_w, er_x+er_w);
done:
  er_x= x;
  er_y= y; 
}

static void eraser_line(int x0, int y0, int x1, int y1)
{
   int dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
   int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
   int err = dx+dy, e2; /* error value e_xy */
   int mx, my;
 
   mx= 0; my= 0;
   for(;;){  /* loop */
      if (mx||my) eraser_point(x0,y0, mx,my);
      if (x0==x1 && y0==y1) break;
      e2 = 2*err;
      mx= 0; my= 0;
      if (e2 >= dy) { err += dy; x0 += sx; mx= sx; } /* e_xy+e_x > 0 */
      if (e2 <= dx) { err += dx; y0 += sy; my= sy; } /* e_xy+e_y < 0 */
   }
}


#include "aathick.c"

static double thickness(double *K, int d1, int d2) { return *K; }

static void drawAAline(int x0,int y0,int x1,int y1,int color)
{
  double T;
  T= 3.2/2;
  draw_varthick_line(x0,y0,x1,y1, (void*) thickness, &T,
                                  (void*) thickness, &T, color);
}


static void begin_cupdate()
{
  ca_uxl= ca_w-1; ca_uyl= ca_h-1;
  ca_uxg= 0;      ca_uyg= 0;
}

static void end_cupdate()
{
  gfx_canvas_update(Wcanvas, ca_uxl, ca_uyl, ca_uxg-ca_uxl+1, ca_uyg-ca_uyl+1);
}

static void canvas_moved(int x,int y)
{
  if (x==ca_x && y==ca_y) return ;
#ifdef REPORTMOVE
    DPRINT(("{'m', %d, %d},\n",x,y))
#endif

  begin_cupdate();
  if (ca_sel==3) 
    eraser_line(ca_x, ca_y, x, y);
  else
  {
    if (ca_pen_width[ca_sel]<=3)
    {
      drawAAline(ca_x, ca_y, x, y, colors[ca_sel]);
    } 
    else
    {
      brush_line(init_poly, ca_shapes[ca_sel], ca_x, ca_y, x, y,colors[ca_sel]);
      init_poly= 1;
    }
  }
  
  end_cupdate();
  ca_x= x;
  ca_y= y;
  ca_mod= 1;
}

void select_pen(int p)
{
  int i;
  static const int widgets[4]= { Wpen1, Wpen2, Wpen3, Wpen4 };
  
  ca_sel= p;
  for(i=0;i<4;i++)
   gfx_set_state(widgets[i], i==p);
}

void paint_eraser(int fg)
{
  int i,dx,dy;
  begin_cupdate();
  for(dx=-er_w;dx<=er_w;dx++)
    for(dy=-er_h;dy<=er_h;dy++)
    if (fg) vline_fg(ca_x+dx, ca_y-er_h, ca_y+er_h, 0xFFffFF);
       else vline_bg(ca_x+dx, ca_y-er_h, ca_y+er_h);
  end_cupdate();
  er_x= ca_x;
  er_y= ca_y;
}

static void paste_sprite_move(gfx_event_t *event)
{
  ca_x= event->x - paste_w/2;
  ca_y= event->y - paste_h/2;
  gfx_sprite_move(Wcanvas, paste_sprite, ca_x, ca_y);
}

static void blend_to_data(int k, unsigned int C)
{
  unsigned int pR,pG,pB,P, cR,cG,cB,cA;

  cA= C>>24;
  if (!cA) return ;
  if (cA==255) goto finish;

  cR= C>>16;     cG= C>>8;     cB= C;
  cR &= 0xff;    cG &= 0xff;   cB &= 0xff;

  P= ca_data[k];
  pR= P>>16;     pG= P>>8;     pB= P;
  pR &= 0xff;    pG &= 0xff;   pB &= 0xff;

  cR= (cR*cA+pR*(255-cA)) / 255;
  cG= (cG*cA+pG*(255-cA)) / 255;
  cB= (cB*cA+pB*(255-cA)) / 255;
  C= (cR<<16) | (cG<<8) | (cB) | (C&TOPBITS);
finish:
  ca_data[k]= C;
}

static void blend_to_sata(int k, unsigned int C)
{
  unsigned int pR,pG,pB,pA, P, cR,cG,cB,cA;

  cA= C>>24;
  if (!cA) return ;
  if (cA==255) goto finish;

  P= ca_sata[k];
  pA= P >> 24;
  pR= P>>16;     pG= P>>8;     pB= P;
  pR &= 0xff;    pG &= 0xff;   pB &= 0xff;

  cR= C>>16;     cG= C>>8;     cB= C;
  cR &= 0xff;    cG &= 0xff;   cB &= 0xff;

  cR= (cR*cA+pR*pA) / (cA+pA);
  cG= (cG*cA+pG*pA) / (cA+pA);
  cB= (cB*cA+pB*pA) / (cA+pA);
  cA= cA + pA;
  if (cA>255) cA= 255;

  C= (cR<<16) | (cG<<8) | (cB) | (cA<<24);
finish:
  ca_sata[k]= C;
}

static void copy_sprite_to_canvas()
{
  int x,y;
  int cw, ch, ox, oy;
  int sx, sy;
  unsigned int P;

  if (!paste_d) return ;

  cw= paste_w; ch= paste_h;
  ox= ca_x;    oy= ca_y;
  sx= 0;       sy= 0;
 
  if (ox<0) { sx-= ox; cw+= ox; ox= 0; }
  if (oy<0) { sy-= oy; ch+= oy; oy= 0; }
  if (ox+cw>ca_w) cw= ca_w-ox;
  if (oy+ch>ca_h) ch= ca_h-oy;
  if (sx+cw>paste_w) cw= paste_w - sx;
  if (sy+ch>paste_h) ch= paste_h - sy;

DPRINT(("pasting image: sx= %d sy= %d ox= %d oy= %d cw= %d ch= %d\n", sx,sy, ox,oy, cw,ch))

  if (cw>0 && ch>0)
    for(y=0;y<ch;y++)
     for(x=0;x<cw;x++)
     {
       int k;
       P= paste_d[(x+sx)+(y+sy)*paste_w];
       k= (x+ox)+(y+oy)*ca_w;
       if (P&0xFF000000)
         { blend_to_data(k,P);
           blend_to_sata(k,P); }
     }

  if (paste_free_data) free(paste_d);
  paste_d= 0;
  ca_mod= 1;
}

static int canvas_paste_event(gfx_event_t *event)
{
  switch(event->type)
  {
  case GFXEV(Wcanvas,BUTTON_PRESS):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    gfx_canvas_begin_report_mouse_position(Wcanvas);
    paste_sprite_move(event);
    gfx_sprite_show(Wcanvas, paste_sprite, 1);
    break;
  case GFXEV(Wcanvas,BUTTON_RELEASE):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    gfx_canvas_end_report_mouse_position(Wcanvas);
    ca_mode= 0;
    copy_sprite_to_canvas();
    gfx_sprite_show(Wcanvas, paste_sprite, 0);
    gfx_sprite_remove(Wcanvas, paste_sprite);
    paste_sprite= -1;
    break;
  case GFXEV(Wcanvas,POINTER_MOVE):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    paste_sprite_move(event);
    break;
  default:
    return 0;
  }
  return 1;
}


int canvas_event(gfx_event_t *event)
{
  switch(event->type)
  {
  case GFXEV(200,TIMEOUT): osd_fadeout(); return 1;
  }
  if (ca_mode=='S') return canvas_selmode_event(event);
  if (ca_mode=='P') return canvas_paste_event(event);
  switch(event->type)
  {
  case GFXEV(Wcanvas,BUTTON_PRESS):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
#ifdef REPORTMOVE
    DPRINT(("/*******BEGIN*******/\n"))
    DPRINT(("{'d', %d, %d },\n",event->x,event->y))
#endif
    gfx_canvas_begin_report_mouse_position(Wcanvas);
    ca_x= event->x; 
    ca_y= event->y;
    if (ca_sel==3) paint_eraser(1);
    init_poly= 0;
    break;
  case GFXEV(Wcanvas,BUTTON_RELEASE):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    canvas_moved(event->x,event->y);
    gfx_canvas_end_report_mouse_position(Wcanvas);
#ifdef REPORTMOVE
    DPRINT(("/*******END*******/\n"))
#endif
    if (ca_sel==3) paint_eraser(0);
    break;
  case GFXEV(Wcanvas,POINTER_MOVE):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    canvas_moved(event->x, event->y);
    break;
  case Wpen1: if (SEQ('T')) break; select_pen(0); break;
  case Wpen2: if (SEQ('P')) break; select_pen(1); break;
  case Wpen3: if (SEQ('R')) break; select_pen(2); break;
  case Wpen4:
    if (SEQ('E')) break; 
    switch(ca_sel)
    {
     case 0: er_w= config.ptitle.ewidth; er_h= config.ptitle.eheight; break;
     case 2: er_w= config.philight.ewidth; er_h= config.philight.eheight; break;
     default: er_w= config.ppen.ewidth; er_h= config.ppen.eheight; break;
    }
    er_w/= 2; er_h /= 2;
    select_pen(3); 
    break;
  default:
    return 0;
  }
  return 1;
}

static void paint_bg(int x,int y,int w,int h)
{
  int xx,yy;
  if (x<0) { w+=x; x= 0; }
  if (y<0) { h+=y; y= 0; }
  if (x+w>=ca_w) w= ca_w-x;
  if (y+h>=ca_h) h= ca_h-y;
  if (w<=0 || h<=0) return ;
  for(yy=y;yy<y+h;yy++)
    for(xx=x;xx<x+w;xx++)
      ca_data[xx+yy*ca_w]= bp_d[ xx%bp_w + (yy%bp_h)*bp_w ];
}

static void copy_data(unsigned int w,unsigned int h, uint32_t *data)
{
  int cw, ch;
  int cx, cy,x,y;

  cw= imin2(w, ca_w); ch= imin2(h, ca_h);  

  if (cw<ca_w) cx= (ca_w-cw)/2;  else cx= 0;
  if (ch<ca_h) cy= (ca_h-ch)/2;  else cy= 0;

  paint_bg(0,0, ca_w, ca_h);

  for(y=0;y<ch;y++)
    for(x=0;x<cw;x++)
    {
      uint32_t D;
      D= data[x+y*w];
                 // FIXME: do proper anti-alias stuff
      if (D&(0xff<<24))
        ca_data[ (x+cx) + (y+cy)*ca_w ]= D;
    }
  gfx_canvas_update(Wcanvas,0,0,ca_w,ca_h);
}




void canvas_set_data(int w,int h,unsigned int *data,unsigned int bg)
{
  uint32_t rw,rh,*rdata;
  ca_mod= 0;
  canvas_set_bg(bg);
  if (w<=ca_w && h<=ca_h) goto noresize;
  if (resize_image_argb(data,w,h,ca_w,ca_h, &rdata, &rw, &rh)) 
    goto noresize;
  copy_data(rw,rh,rdata);
  free(rdata);
  return ;
noresize:
  copy_data(w,h,data);
}

void canvas_clear()
{
  ca_mod= 0;
  memset(ca_sata, 0, ca_w*ca_h*sizeof(ca_sata[0]));
  paint_bg(0, 0, ca_w, ca_h);
  gfx_canvas_update(Wcanvas,0,0,ca_w, ca_h);
}

  // we always deal with saved data externally, so the data
  // part 
void canvas_get_data(uint32_t *Rw, uint32_t *Rh, uint32_t **Rd)
{
  *Rw= ca_w;
  *Rh= ca_h;
  *Rd= ca_sata;
}

// this is the actual displayed data part, with the background
// already painted.
void canvas_get_fdata(uint32_t *Rw, uint32_t *Rh, uint32_t **Rd)
{
  *Rw= ca_w;
  *Rh= ca_h;
  *Rd= ca_data;
}


int canvas_modified()
{
  return ca_mod;
}

void canvas_set_modified() { ca_mod=1; }



void canvas_begin_select()
{
  ca_mode= 'S';
  oplogl= 0;
}

static void sel_pixel(int x,int y, int color)
{
  unsigned int k;
  if (x<0 || x>=ca_w) return ;
  if (y<0 || y>=ca_h) return ;

  k= x+y*ca_w;
  {
    if (!(ca_data[k]&TOPBITS))  // only mark background pixels
       ca_data[k] = color;
  }

  ca_uxl= imin2(ca_uxl, x);  ca_uxg= imax2(ca_uxg, x);
  ca_uyl= imin2(ca_uyl, y);  ca_uyg= imax2(ca_uyg, y);
}




void draw_line(int x0, int y0, int x1, int y1, int color)
{
  int sx, sy, dx, dy, x, y;
  int thres;
  int e;
  int Ediag, Esqur;

  sx= (x0>x1) ? -1 : 1;
  sy= (y0>y1) ? -1 : 1;
  dx= (x1-x0)*sx;
  dy= (y1-y0)*sy;

  e= 0;
  if (dx>=dy)
  {
    Ediag= -2 * dx;
    Esqur=  2 * dy;
    thres= dx - 2*dy;
    y= y0;
    for(x=x0;x!=x1+sx;x+=sx)
    {
      sel_pixel(x,y,color);
      if (e > thres) { y+= sy; e+= Ediag; }
      e+= Esqur;
    }
  }
  else
  {
    Ediag= -2 * dy;
    Esqur=  2 * dx;
    thres= dy - 2*dx;
    x= x0;
    for(y=y0;y!=y1+sy;y+=sy)
    {
      sel_pixel(x,y,color);
      if (e > thres) { x+= sx; e+= Ediag; }
      e+= Esqur;
    }
  }
}


static void canvas_selmode_moved(int ex,int ey)
{
  static int z=1;
  if (ex==ca_x && ey==ca_y) return ;
  ca_uxl= ca_w-1; ca_uyl= ca_h-1;
  ca_uxg= 0;      ca_uyg= 0;

  draw_line(ca_x, ca_y, ex, ey, bp_mark);
  gfx_canvas_update(Wcanvas, ca_uxl, ca_uyl, ca_uxg-ca_uxl+1, ca_uyg-ca_uyl+1);

  if (oplogl==oplogs) 
  {
     int S;
     oplogs += oplogs / 2 + 100;
     S= sizeof(oplog[0])*oplogs*2;
     if (!oplog) oplog= malloc(S);
            else oplog= realloc(oplog, S);
  } 

  oplog[2*oplogl]= ex; oplog[2*oplogl+1]= ey;
  oplogl++;

  ca_x= ex;
  ca_y= ey;
  ca_mod= 1;
}

static int linear_delta(int d1x,int d1y,int d2x,int d2y)
{
  int K,M;
  if (d1x==0) return d2x==0;
  if (d1y==0) return d2y==0;
  if (d2x==0 || d2y==0) return 0;

  K= d1x*d2y;
  M= d2x*d1y;
  return (K-M==0) || (K+M==0);
}

static int linear_triple(int i,int j,int k)
{
  int d1x, d1y, d2x, d2y;
  d1x= oplog[2*j+0]-oplog[2*i+0]; d1y= oplog[2*j+1]-oplog[2*i+1];
  d2x= oplog[2*k+0]-oplog[2*j+0]; d2y= oplog[2*k+1]-oplog[2*j+1];
  return linear_delta(d1x,d1y,d2x,d2y);
}

static  void copy_point(int d,int s)
{
  oplog[2*d+0]= oplog[2*s+0];
  oplog[2*d+1]= oplog[2*s+1];
}

static void filter_linear_triples()
{
  int i;
  for(i=0;i+2<oplogl;i++)
   if (linear_triple(i,i+1,i+2))
   {
     int m;
     for(m=i+1;m<oplogl-1;m++)
       copy_point(m,m+1);
     oplogl--;
     i--;
   }

#if 0
 { FILE *ff;
   ff= fopen("poly.txt","wb");
   if (!ff) return ;
 for(i=0;i<oplogl;i++)
  fprintf(ff,"%d %d\n", oplog[2*i+0], oplog[2*i+1]);
 fprintf(ff,"-1 -1\n");
 fclose(ff); }
#endif
}

static const char *pattern[]=
{
  "..X...X.",
  "...X...X",
  "...X...X",
  "X...X...",
  "X...X...",
  "...X...X",
  "..X...X.",
  ".X...X..",
};

static int patW= 8, patH=8;

static void filpixel(int x,int y)
{
  unsigned int P;
  if (x<0 || y<0 || x>=ca_w || y>=ca_h) return ;
  ca_uxl= imin2(ca_uxl, x);  ca_uxg= imax2(ca_uxg, x);
  ca_uyl= imin2(ca_uyl, y);  ca_uyg= imax2(ca_uyg, y);

  P= ca_sata[x+y*ca_w];
  if ((P&0xFF000000) && x>=sel_mix && x<=sel_max
                     && y>=sel_miy && y<=sel_may)
   sel_d[ (x-sel_mix) + (y-sel_miy)*sel_w ]= P;

  if (pattern[y%patH][x%patW]!='X') return ;
  if (P& (0xFF<<24)) return ;
  ca_data[x+y*ca_w] = bp_mark;
}

#define FILLPOLY_PROTO static void fill_polygon(int npoints, int *points)
#define FILLPOLY_PIXEL(x,y) filpixel(x,y)
#include "fillpoly.c"

static int opsel_scr()
{
  gfx_event_t event;
  int R;
  gfx_SELCLIPOP();
  R= 0;
  while(!R)
   if (gfx_event(NULL,&event))
    switch(event.type)
    {
    case WsoERASE:   R= 'E'; break;
    case GFXEV_QUIT:
    case WsoCANCEL:  R= 'Q'; break; 
    case WsoCUT:     R= 'U'; break;
    case WsoCOPY:    R= 'C'; break;
    case WsoEXPORT:  R= 'X'; break;
    }
  gfx_leave_screen();
  return R;
}


static void make_selection()
{
  int i,x,y;
  int sop;

  if (oplogl<3) return ;
  filter_linear_triples();
  if (oplogl<3) return ;

  sel_mix= sel_max= oplog[0];
  sel_miy= sel_may= oplog[1];
  for(i=0;i<oplogl;i++)
  {
    sel_mix= imin2(sel_mix, oplog[2*i+0]);
    sel_max= imax2(sel_max, oplog[2*i+0]);
    sel_miy= imin2(sel_miy, oplog[2*i+1]);
    sel_may= imax2(sel_may, oplog[2*i+1]);
  }
  sel_w= sel_max - sel_mix + 1;
  sel_h= sel_may - sel_miy + 1;
  sel_d= malloc(sel_w*sel_h*sizeof(sel_d[0]));
  for(i=0;i<sel_w*sel_h;i++)
    sel_d[i]= 0;

  ca_uxl= ca_w-1; ca_uyl= ca_h-1;
  ca_uxg= 0;      ca_uyg= 0;
  fill_polygon(oplogl, oplog);
  gfx_canvas_update(Wcanvas, ca_uxl, ca_uyl, ca_uxg-ca_uxl+1, ca_uyg-ca_uyl+1);

    // Erase, Quit, cUt, Copy, eXport
  sop= opsel_scr();
  if (sop=='E' || sop=='U')
  {
      // remove data from canvas
     for(y=0;y<sel_h;y++)
       for(x=0;x<sel_w;x++) 
         if (sel_d[x+y*sel_w]&0xFF000000)
         {
            ca_sata[(x+sel_mix)+(y+sel_miy)*ca_w]= 0;
            ca_data[(x+sel_mix)+(y+sel_miy)*ca_w]= 
               bp_d[ (x+sel_mix)%bp_w + ((y+sel_miy)%bp_h)*bp_w ];
         }
  }

    // remove the selection display
  for(y=ca_uyl;y<=ca_uyg;y++)
    for(x=ca_uxl;x<=ca_uxg;x++)
      if (!(ca_data[x+y*ca_w]&0xFF000000))
         ca_data[x+y*ca_w]= bp_d[ x%bp_w + (y%bp_h)*bp_w ];

  gfx_canvas_update(Wcanvas, ca_uxl, ca_uyl, ca_uxg-ca_uxl+1, ca_uyg-ca_uyl+1);

  switch(sop)
  {
  case 'U': case 'C': save_selected_data(); break;
  case 'X':           export_image(sel_w,sel_h,sel_d, bp_id); // fallthru
  case 'E': case 'Q': free(sel_d); sel_d= NULL; break;
  }
}

static int canvas_selmode_event(gfx_event_t *event)
{
  switch(event->type)
  {
  case GFXEV(Wcanvas,BUTTON_PRESS):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
        // multiple presses occur sometimes, I don't know why
    if (oplogl)
    {
      canvas_selmode_moved(event->x,event->y);
      break;
    }
    gfx_canvas_begin_report_mouse_position(Wcanvas);
    ca_x= event->x; 
    ca_y= event->y;
    break;
  case GFXEV(Wcanvas,BUTTON_RELEASE):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    canvas_selmode_moved(event->x,event->y);
    gfx_canvas_end_report_mouse_position(Wcanvas);
    ca_mode= 0;
    make_selection();
    break;
  case GFXEV(Wcanvas,POINTER_MOVE):
#ifdef GFX_ANDROID
    if (event->key>FINGER_THRESHOLD) break;
#endif
    canvas_selmode_moved(event->x,event->y);
    break;
  default:
    return 0;
  }
  return 1;
}

#define CLI_W 200
#define CLI_H 267
static unsigned int cli_d[CLI_W*CLI_H];

static void insert_clip_thumb
  (unsigned int w, unsigned int h, unsigned int *data,unsigned int bg)
{
  int i;
  unsigned int *rdata, rw, rh;
  unsigned int bW, bH, *bD, bM;
  int ox, oy, cw, ch,x,y;

  gen_pattern(bg, &bW, &bH, &bD, &bM); 
  rdata= NULL;

  for(y=0;y<CLI_H;y++)
    for(x=0;x<CLI_W;x++)
      cli_d[x+y*CLI_W]= bD[ (x%bW) + (y%bH)*bW ];

  if ((w<=CLI_W && h<=CLI_H) || 
      resize_image_argb(data,w,h,CLI_W,CLI_H,&rdata,&rw, &rh))
    { rdata= data; rw= w; rh= h; }

  cw= imin2( rw, CLI_W );
  ch= imin2( rh, CLI_H );
  ox= (CLI_W-cw)/2;
  oy= (CLI_H-ch)/2;
  for(y=0;y<ch;y++)
    for(x=0;x<cw;x++)
    {
      int k;
      unsigned int pA, pR, pG, pB, P;
      unsigned int     bR, bG, bB, B;
      k= (x+ox)+(y+oy)*CLI_W;
      P= rdata[x+y*rw];
      pA= P>>24; pR= P>>16; pG= P>>8;  pB= P;
                 pR&= 0xff; pG&= 0xff; pB&= 0xff; 
      if (pA==0) continue;
      else if (pA==255) { cli_d[k]= P; continue; }
      B= cli_d[k];
                 bR= B>>16; bG= B>>8;  bB= B;
                 bR&= 0xff; bG&= 0xff; bB&= 0xff; 
      pR= (pR*pA+bR*(255-pA))/255;
      pG= (pG*pA+bG*(255-pA))/255;
      pB= (pB*pA+bB*(255-pA))/255;
      P= (pR<<16) | (pG<<8) | (pB);
      cli_d[k]= P;
    }
  if (rdata && rdata!=data) free(rdata);
  gfx_iconlist_insert(WcbLIST, -1, "", CLI_W, CLI_H, cli_d, 0, NULL);
}


void insert_from_clipboard()
{
  int i;
  int S,itemno;
  gfx_event_t event;

  gfx_CLIPS();
  gfx_iconlist_updates(WcbLIST, 0);
  for(i=0;i<NCLIPS;i++)
   if (clipboard[i].data)
      insert_clip_thumb(clipboard[i].w, clipboard[i].h, 
                        clipboard[i].data, clipboard[i].bg);
   else
      break;
  gfx_iconlist_updates(WcbLIST, 1);
  S= 1; itemno= -1;
  while(S)
  {
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case WcbCLOSE: S=0; break;
    case WcbLIST: itemno= event.key; S= 0; break;
    }
  }
  gfx_leave_screen();
  if (itemno==-1) return ;

  ca_mode= 'P';
  paste_w= clipboard[itemno].w;
  paste_h= clipboard[itemno].h;
  paste_d= clipboard[itemno].data;
  paste_free_data= 0;
  paste_sprite= gfx_sprite_new(Wcanvas,'S', paste_w, paste_h, paste_d);
}


void canvas_insert_image(unsigned int w, unsigned int h, unsigned int *d)
{
  ca_mode= 'P';
  paste_w= w;
  paste_h= h;
  paste_d= d;
  paste_free_data= 1;
  paste_sprite= gfx_sprite_new(Wcanvas,'S', paste_w, paste_h, paste_d);
}

static int osd_w,osd_h; static unsigned int *osd_d;
static int osd_x, osd_y, osd_sprite=-1;
static int osd_count, osd_delay;

#define OSD_STEPS  10
#define COL_INT (TOPBITS | 0)
#define COL_BOR (TOPBITS | 0xff8888cc)
#define COL_MSG (TOPBITS | 0xff88cccc)
#define PAD 5

static void blend_color(unsigned int *dd, unsigned char V, unsigned int C)
{
  int cR, cG, cB;
  int dR, dG, dB;
  unsigned int D;
  
  D= *dd;

  dR= (D>>16) & 0xff; dG= (D>>8) & 0xff; dB= (D>>0) & 0xff;
  cR= (C>>16) & 0xff; cG= (C>>8) & 0xff; cB= (C>>0) & 0xff;

  dR= (cR*V + dR*(255-V)) / 255;
  dG= (cG*V + dG*(255-V)) / 255;
  dB= (cB*V + dB*(255-V)) / 255;

  *dd= TOPBITS | (dR<<16) | (dG<<8) | dB;
}

static void display_osd_sprite()
{
  gfx_pause();
  if (osd_sprite!=-1)
    gfx_sprite_remove(Wcanvas, osd_sprite);
  osd_sprite= gfx_sprite_new(Wcanvas, 'S', osd_w, osd_h, osd_d);
  gfx_sprite_move(Wcanvas, osd_sprite, osd_x, osd_y);
  gfx_resume();
  gfx_sprite_show(Wcanvas, osd_sprite, 1);
  gfx_timer_start(200, osd_delay);
}

static void osd_fadeout()
{
  int x,y;
  int V;
  if (osd_count==OSD_STEPS) { osd_stop(); return ; }
  osd_count++;
  V= 255/OSD_STEPS;
  for(y=0;y<osd_h;y++)
  { if (y+osd_y<0) continue;
    if (y+osd_y>=ca_h) break;
    for(x=0;x<osd_w;x++)
    { if (x+osd_x<0) continue;
      if (x+osd_x>=ca_w) break; 
      blend_color(osd_d+x+y*osd_w, V,
                         ca_data[(x+osd_x)+(y+osd_y)*ca_w]);
    }
  }
  display_osd_sprite();
}


void canvas_print_tempo(int ms, const char *fmt, ...)
{
  va_list ap;
  char buf[2000];
  int x,y,tw,th;
  unsigned char *data;
  va_start(ap,fmt);
  vsnprintf(buf,sizeof(buf)-1,fmt,ap);
  va_end(ap);
  buf[sizeof(buf)-1]= 0;
  
  gfx_text_data(1,buf,&tw,&th,NULL,&data);
  osd_w= tw+2*PAD; osd_h= th+2*PAD;
  osd_d= malloc(sizeof(osd_d)*osd_w*osd_h);

  for(y=1;y<osd_h-1;y++)
    for(x=1;x<osd_w-1;x++)
      osd_d[x+y*osd_w]= COL_INT;
  for(x=0;x<osd_w;x++)
    osd_d[x+0]= osd_d[x+(osd_h-1)*osd_w]= COL_BOR;
  for(y=1;y<osd_h-1;y++)
    osd_d[0+y*osd_w]= osd_d[ (osd_w-1)+y*osd_w ]= COL_BOR;

  for(y=0;y<th;y++)
    for(x=0;x<tw;x++)
      blend_color(osd_d+(x+PAD)+(y+PAD)*osd_w, data[x+y*tw], COL_MSG);

  osd_x= (ca_w-osd_w)/2;
  osd_y= (ca_h-osd_h)-2;
  osd_sprite= -1;
  osd_count= 0;
  gfx_timer_stop(200);
  osd_delay= ms / OSD_STEPS;
  display_osd_sprite();
}

static void osd_stop()
{
  if (osd_d) free(osd_d);
  osd_d= NULL;
  gfx_timer_stop(200);
  if (osd_sprite!=-1) gfx_sprite_remove(Wcanvas,osd_sprite);
  osd_sprite= -1;
}

 // this function is called after the data has been read into 
 // the canvas buffer. All I need to do is to paint the background
 // areas

static void sata_to_data(int x,int y, int w, int h)
{
  unsigned int k;
  unsigned int P, C;
  unsigned int pR,pG,pB, cR, cG, cB, value;
  int xp, yp;

  if (x<0) { w+= x; x= 0; }
  if (y<0) { h+= x; y= 0; }
  if (x+w>ca_w) w= ca_w-x; 
  if (y+h>ca_h) h= ca_h-y;
  if (w<=0 || h<=0) return ;

  for(yp=0;yp<h;yp++)
  for(xp=0;xp<w;xp++)
   {
     k= (x+xp) + (y+yp)*ca_w;
     P= ca_sata[k];

     value= P >> 24;
     if (value==255) { ca_data[k]= P; continue; }

     C= bp_d[ (x+xp)%bp_w + ((y+yp)%bp_h)*bp_w ];
     if (value==0)   { ca_data[k]= C; continue; }

     pR= P>>16;     pG= P>>8;     pB= P;
     pR &= 0xff;    pG &= 0xff;   pB &= 0xff;

     cR= C>>16;     cG= C>>8;     cB= C;
     cR &= 0xff;    cG &= 0xff;   cB &= 0xff;
     pR= (pR*value+cR*(255-value)) / 255;
     pG= (pG*value+cG*(255-value)) / 255;
     pB= (pB*value+cB*(255-value)) / 255;

     ca_data[k]= (pR<<16) | (pG<<8) | (pB) | TOPBITS;
   }
}

void canvas_update_data(int uw, int uh, unsigned int bg)
{
  int bw,x,y;
  ca_mod= 0;
  bw= ca_w;
  x= 0;
  canvas_set_bg(bg);
  if (uw<ca_w)
  {
    int x;
    paint_bg(0,0, (ca_w-uw)/2, ca_h);
    x= (ca_w-uw)/2 + uw;
    paint_bg( x, 0, ca_w-x, ca_h);
    x= (ca_w-uw)/2; bw= uw;
  }
  if (uh<ca_h)
  {
    paint_bg(x,0, bw, (ca_h-uh)/2);
    y= (ca_h-uh)/2 + uh;
    paint_bg(x,y, bw, ca_h-y);
  }
  sata_to_data((ca_w-uw)/2, (ca_h-uh)/2, uw, uh);
  gfx_canvas_update(Wcanvas,0,0,ca_w, ca_h);
}

void canvas_set_bg(unsigned int bg)
{
  bp_id= bg;
  gen_pattern(bg, &bp_w, &bp_h, &bp_d, &bp_mark); 
}

