
#ifndef FILLPOLY_RETURN 
#define FILLPOLY_RETURN
#endif

#ifndef FILLPOLY_PIXELARGS
#define FILLPOLY_PIXELARGS
#endif

#ifndef FILLPOLY_PIXELPARAMS
#define FILLPOLY_PIXELPARAMS
#endif

#ifndef FILLPOLY_INIT
#define FILLPOLY_INIT
#endif

typedef struct
{
   // start-end points
  int x0, x1, y0, y1; 
  int ex0, ex1; // extended x for the start and end points
  int x;
  int minx,maxx;

  int xline;                       // is 1 if the line is X based
  int xstep;
  int error,dx,dy;                 // bresenham values
  int e_diag, e_square, threshold; 

#if defined(FILLPOLY_DEBUG_PRINT) || defined(FILLPOLY_DEBUG_GFX) 
  int ident;
#endif
} edge_t;

typedef struct
{
  edge_t **data;
  int head, len, size;
} equeue_t;

#if defined(FILLPOLY_MAXMINFUNC)
static int imin2(int a,int b) { return a<b? a:b; }
static int imax2(int a,int b) { return a>b? a:b; }
#endif

static void push_edge(equeue_t *Q, edge_t *E)
{
  Q->data[(Q->head+Q->len)%Q->size]= E;
  Q->len++;
}

static edge_t* get_edge(equeue_t *Q)
{
  edge_t *E;
  E= Q->data[Q->head];
  Q->head= (Q->head+1)%Q->size;
  Q->len--;
  return E;
}

static edge_t* top_edge(equeue_t *Q)
{
  return Q->data[Q->head];
}

static equeue_t *make_queue(int ne)
{
  equeue_t *Q;
  Q= malloc(sizeof(*Q));
  Q->size= ne;
  Q->len= Q->head= 0;
  Q->data= malloc(sizeof(Q->data[0])*ne);
  return Q;
}

static void free_queue(equeue_t *Q)
{
  free(Q->data);
  free(Q);
}

static int sgn(int x) { return x<0 ? -1:1; }

static int compar_edge(const void *A, const void *B)
{
  const edge_t *E1= A;
  const edge_t *E2= B;
  const edge_t *t;
  int R,swap;
  int dxA,dxB,dyA,dyB;
  int s;

  R= E1->y0 - E2->y0;
  if (R) return R;
  if (E2->y0==E2->y1) { swap= 1; goto horz; }
  if (E1->y0==E1->y1) { swap= -1; t= E1; E1= E2; E2= t; goto horz; }
  R= E1->x0 - E2->x0;
  if (R) return R;

  dxB= E2->x1-E2->x0;
  dxA= E1->x1-E1->x0;
  if (sgn(dxA)!=sgn(dxB)) return E1->x1-E2->x1;

  s= sgn(dxA);
  dyA= E1->y1-E1->y0;
  dyB= E2->y1-E2->y0;

  dxA *= s; dxB *= s;
  if (dxA*dyB/dyA<dxB) return -s;
  return s;
horz:
  // if two horizontals are compared, we do it 
  // according to their x0
  if (E1->y0==E1->y1) return (E1->x0-E2->x0)*swap;

  if (E1->x0==E2->x0 || E1->x1==E2->x0) return -1*swap;
  if (E1->x0==E2->x1 || E1->x1==E2->x1) return 1*swap;
  R= (E1->x0-E2->x0)*swap;
  if (R) return R;
  R= (E1->x1-E2->x1)*swap;
  return R;
}


static void make_edges
   (unsigned int npoints, int *points,
    int *Rnedges,edge_t **Redges)
{
  edge_t *edges;
  int nedges;
  int i;

  edges= calloc(npoints,sizeof(edges[0]));
  nedges= 0;

#define XCOOR(k)   points[ 2*((k)%npoints) + 0 ]
#define YCOOR(k)   points[ 2*((k)%npoints) + 1 ]
  for(i=0;i<npoints;i++)
  {
    int x0,x1,y0,y1,ex0,ex1;
    x0= XCOOR(i);
    y0= YCOOR(i);
    x1= XCOOR(i+1);
    y1= YCOOR(i+1);

    if (y1==y0) continue;

      // followed by a horizontal
    if ( YCOOR(i+2)==y1 ) ex1= XCOOR(i+2); else ex1= x1;
      // preceeded by a horizontal
    if ( YCOOR(i-1+npoints)==y0 ) ex0= XCOOR(i-1+npoints); else ex0= x0;

    if (y1<y0)
    {
      edges[nedges].x0= x1; edges[nedges].y0= y1; edges[nedges].ex0= ex1;
      edges[nedges].x1= x0; edges[nedges].y1= y0; edges[nedges].ex1= ex0;
    }
    else
    {
      edges[nedges].x0= x0; edges[nedges].y0= y0; edges[nedges].ex0= ex0;
      edges[nedges].x1= x1; edges[nedges].y1= y1; edges[nedges].ex1= ex1;
    }
    nedges++;
  }

  *Redges= edges;
  *Rnedges= nedges;

#ifdef FILLPOLY_DEBUG_PRINT
  for(i=0;i<nedges;i++) edges[i].ident= i;
#endif

  qsort(edges,nedges,sizeof(edges[0]),compar_edge);

#ifdef FILLPOLY_DEBUG_PRINT
for(i=0;i<nedges;i++)
{
printf("%3d: [%4d (%4d),%4d] -> [%4d (%4d),%4d]\n",
      edges[i].ident,
          edges[i].x0, edges[i].ex0, edges[i].y0,
          edges[i].x1, edges[i].ex1, edges[i].y1);
}
#endif

#ifdef FILLPOLY_DEBUG_GFX
  for(i=0;i<nedges;i++) DBG_ident_line((edges[i].x0+edges[i].x1)/2,
                                       (edges[i].y0+edges[i].y1)/2, 
                                       edges[i].ident);
#endif
#undef XCOOR
#undef YCOOR
}


static void advance_edge_xline(edge_t *E,int y)
{
  int lox;
  E->x= lox= E->x + E->xstep;
  while(E->error<E->threshold && E->x!=E->x1+E->xstep)
  {
    E->error+= E->e_square;
    E->x+= E->xstep;
  } 
  E->error+= E->e_diag;
  E->x-= E->xstep;
  E->minx= imin2(E->x,lox);
  E->maxx= imax2(E->x,lox);

  if (y==E->y1-1) { E->minx= imin2(E->minx, E->ex1);
                    E->maxx= imax2(E->maxx, E->ex1); }
}

static void advance_edge_yline(edge_t *E,int y)
{
  if (E->error>=E->threshold)
  {
    E->error+= E->e_diag;
    E->x+= E->xstep;
  }
  E->minx= E->maxx= E->x;
  E->error+= E->e_square;

  if (y==E->y1-1) { E->minx= imin2(E->x, E->ex1);
                    E->maxx= imax2(E->x, E->ex1); }
}

static void init_edge_xline(edge_t *E)
{
  E->error= 0;
  E->threshold= E->dx-2*E->dy;
  E->e_diag= -2*E->dx;
  E->e_square= 2*E->dy;

  E->x= E->x0;
  while(E->error<E->threshold && E->x!=E->x1+E->xstep)
  {
    E->x+= E->xstep;
    E->error += E->e_square;
  }  
  E->x-= E->xstep;
  E->error+= E->e_diag;
  E->xline= 1;
  
  E->minx= imin2(E->ex0, imin2(E->x,E->x0));
  E->maxx= imax2(E->ex0, imax2(E->x,E->x0));
}

static void init_edge_yline(edge_t *E)
{
  E->minx= E->maxx= E->x= E->x0;
  E->error= 0;
  E->threshold= E->dy-2*E->dx;
  E->e_diag= -2*E->dy;
  E->e_square= 2*E->dx;
  E->xline= 0;

  E->minx= imin2(E->ex0, E->minx);
  E->maxx= imax2(E->ex0, E->maxx);
}

static void advance_edge(edge_t *E,int y)
{
  if (E->xline) advance_edge_xline(E,y); 
          else  advance_edge_yline(E,y);
}

static void init_edge(edge_t *E)
{
  E->dx= E->x1 - E->x0;
  E->dy= E->y1 - E->y0;

  if (E->dx<0) { E->dx= -E->dx; E->xstep= -1; }
          else { E->xstep= 1; }

  if (E->dx>E->dy) init_edge_xline(E);
              else init_edge_yline(E);
}

static void findmax(edge_t *edges,int nedges,int *Rmaxy,int *Rminx)
{
  int i,maxy,minx;
  maxy= edges[0].y1;
  minx= imin2(edges[0].x0,edges[0].x1);
  for(i=1;i<nedges;i++) { minx=imin2(minx,imin2(edges[i].x0,edges[i].x1));
                          maxy= imax2(edges[i].y1,maxy); }
  *Rminx= minx;
  *Rmaxy= maxy;
}

static void output_span
   (FILLPOLY_PIXELPARAMS int y, int x0, int x1,int *Rmax)
{
  int x;
  x0= imax2(*Rmax+1, x0);
  for(x=x0;x<=x1;x++)
    FILLPOLY_PIXEL(x,y);
  *Rmax= x1;
}

static int edges_join(edge_t *P,edge_t *C)
{
  return (C->x0==P->x1)  || 
         (C->ex0==P->x1) ||
         (C->x0==P->ex1) ||
         (C->ex0==P->ex1);
}

FILLPOLY_PROTO
{
  edge_t *edges,*E,*C,*P;
  equeue_t *Q;
  int nedges;
  int minx;
  int lastx;
  int i, miny, maxy;
  int candi;
  int y;
  int QL;
  int have_left, left;

  left= 0; // shut up gcc -O3

  FILLPOLY_INIT

  make_edges(npoints, points, &nedges, &edges);

  miny= edges[0].y0;
  findmax(edges,nedges,&maxy,&minx);

  Q= make_queue(nedges);
  candi= 0;

  for(i=0;i<nedges;i++) init_edge(edges+i);

#ifdef FILLPOLY_DEBUG_PRINT 
  printf("array= "); for(i=0;i<nedges;i++) printf("%d.",edges[i].ident); printf("\n");
#endif

  y= miny-1;
scanline_loop:
  lastx= minx-1;
  y++;
  if (y>maxy) goto scanline_done;

  have_left= 0;
  QL= Q->len;

inner_loop:
  if (candi<nedges && edges[candi].y0==y) C= edges+candi;
                                     else C= NULL; 
  if (QL)  P= top_edge(Q);
      else P= NULL;

  switch ((P?2:0) | (C?1:0))
  {
   case 3: if (y==P->y1 && y==C->y0 && edges_join(P,C))
           {
             E=C;
             C->minx= imin2(C->minx,P->minx);
             C->maxx= imax2(C->maxx,P->maxx);
             get_edge(Q); QL--;
             break;
           }
           if (C->x<P->x) E= C; else E= P;
           break;
   case 2: E= P; break;
   case 1: E= C; break;
   case 0: goto scanline_loop;
  }

#ifdef FILLPOLY_DEBUG_PRINT
  if (E==C) { printf("selected candidate %d\n", C->ident);
              if (P) printf("   queue element was %d\n", P->ident); }
#endif


  if (have_left) { output_span(FILLPOLY_PIXELARGS y, left, E->maxx,&lastx);
                   have_left= 0;}
            else { left= E->minx; have_left= 1; }

  if (E==C) candi++;
       else { get_edge(Q); QL--; }

  if (y<E->y1) { advance_edge(E, y); push_edge(Q,E); }

  goto inner_loop;

scanline_done:
  free_queue(Q);
  free(edges);

  FILLPOLY_RETURN 
}
