#include <stdlib.h>
#include <stdio.h>
#include "enviro.h"
#include "genpat.h"

#define unt unsigned int
#define inchinmm  25.4

/***********************************************************************
 This file generates the background patterns. The patterns are generated
 on the fly since they are so easy to do. There is no need to store them
 on disk.

 Each pattern is generated only once, then remembered.
 ***********************************************************************/

static void ialloc(unt w, unt h, unt *Rw, unt *Rh, unt **Rd, unt **Pd)
{
  unt *D;
  D= malloc(sizeof(*D)*w*h);
  *Rd= *Pd= D;
  *Rw= w;
  *Rh= h;
}

static void ifill(unt w, unt h, unt *d, unt color)
{
  unt i;
  for(i=0;i<w*h;i++) d[i]= color;
}

static void gen_10mmLINES
  (unt *Rw, unt *Rh, unt **Rd, unt bg, unt grid)
{
  unt w,h,*d,x;
  w= 2; 
  h=  10 * dpi / inchinmm;
  ialloc(w,h, Rw, Rh, Rd, &d);
  ifill(w,h,d, bg);
  for(x=0;x<w;x++) d[x+(h-1)*w]= grid;
}

static void gen_5mmGRID
   (unt *Rw, unt *Rh, unt **Rd, unt bg, unt grid)
{
  unt w,h,*d,x,y;
  w= h=  5 * dpi / inchinmm;
  ialloc(w,h, Rw, Rh, Rd, &d);
  ifill(w,h,d, bg);
  for(x=0;x<w;x++) d[x+0*w]= grid;
  for(y=0;y<h;y++) d[0+y*w]= grid;
}

static void gen_5mmblack
   (unt *Rw, unt *Rh, unt **Rd, unt *Rbg)
{
  gen_5mmGRID(Rw, Rh, Rd, 0x00, 0x222222);
  *Rbg= 0xFFffFF;
}

static void gen_5mmwhite
   (unt *Rw, unt *Rh, unt **Rd, unt *Rbg)
{
  gen_5mmGRID(Rw, Rh, Rd, 0xffFFff, 0xddDDdd);
  *Rbg= 0x00;
}

static void gen_5mmyellow
   (unt *Rw, unt *Rh, unt **Rd, unt *Rbg)
{
  gen_5mmGRID(Rw, Rh, Rd, 0xa4ad49, 0xb6bf51);
  *Rbg= 0x00;
}

static void gen_line10mmblack
   (unt *Rw, unt *Rh, unt **Rd, unt *Rbg)
{
  gen_10mmLINES(Rw, Rh, Rd, 0x00, 0x222222);
  *Rbg= 0xFFffFF;
}

static void gen_line10mmwhite
   (unt *Rw, unt *Rh, unt **Rd, unt *Rbg)
{
  gen_10mmLINES(Rw, Rh, Rd, 0xffFFff, 0xddDDdd);
  *Rbg= 0x00;
}

static void gen_line10mmyellow
   (unt *Rw, unt *Rh, unt **Rd, unt *Rbg)
{
  gen_10mmLINES(Rw, Rh, Rd, 0xa4ad49, 0xb6bf51);
  *Rbg= 0x00;
}

typedef struct
{
  char *name;
  void (*gen)(unt *, unt *, unt **,unt *);
  unt w, h, *d, bg;
} pat_t;

static pat_t patterns[]= 
{
  { "5 mm", gen_5mmblack }, 
  { "5 mm", gen_5mmwhite },
  { "5 mm", gen_5mmyellow },
  { "10 mm", gen_line10mmblack },
  { "10 mm", gen_line10mmwhite },
  { "10 mm", gen_line10mmyellow }
};

void gen_pattern
  (unsigned int pat, unt *Rw, unt *Rh, unt **Rd,unt *Rbg)
{
  int sz;
  pat_t *P;
  if (pat & 0xff000000)
  {
    pat &= ~0xFF000000;
    sz= sizeof(patterns)/sizeof(patterns[0]);
    if (pat>=sz) pat= sz-1;
    P= patterns+pat;
    if (!P->d) (*P->gen)(&P->w, &P->h, &P->d, &P->bg);
    *Rd= P->d; *Rw= P->w; *Rh= P->h; *Rbg= P->bg;
  }
  else
  {
    static unt buf;
    buf= pat;
    *Rw= 1; *Rh= 1; *Rd= &buf; 
    *Rbg= pat ^ 0xFFffFF;
  }
}

void get_pattern_info
  (unsigned int pat, unt w, unt h, unt *d,
    char **name)
{
  int sz;
  int y,x;
  unsigned int pw, ph, *pd, pb;
  sz= sizeof(patterns)/sizeof(patterns[0]);
  if (pat>=sz) pat= sz-1;
  gen_pattern( pat | 0xff000000, &pw, &ph, &pd, &pb);
  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      d[x+y*w]= pd[ (x%pw) + (y%ph)*pw ];
  *name= patterns[pat].name;
}

int get_npatterns() { return sizeof(patterns)/sizeof(patterns[0]); }

