typedef struct
      {
  int   type;
  int   x, y;
  int   modifiers;
  int   button;
  int   key;
  char *string;
  void *usr;
  int wide;
  double flo0, flo1;
  unsigned long time;
      }
    gfx_event_t;

enum
{
  GFXEV_BUTTON_PRESS=1, GFXEV_BUTTON_RELEASE, GFXEV_DOUBLE_CLICK,
  GFXEV_KEY_PRESS,      GFXEV_KEY_RELEASE,
  GFXEV_POINTER_MOVE,   
  GFXEV_READY_READ, GFXEV_READY_WRITE, GFXEV_READY_EXCEPT,
  GFXEV_QUIT, GFXEV_SAVE_STATE, GFXEV_JAVARESULT, GFXEV_TIMEOUT,
  GFXEV_LAST,
};

#define GFXEV(X,Y) (((X)<<8)|(GFXEV_##Y)|(1<<31))

#define GFXRESERVED 24

/* The event routine returns an event from the queue. It waits a maximum
   of 'msecs' milliseconds for the event. If msecs is zero, it waits
   until an event is received.

   The return value is one if an event is returned. It's zero if the
   return was due to timeout.

   In case of an event, *msecs is updated to reflect the remaining
   wait time.  */

int gfx_event(unsigned int *msecs,gfx_event_t *Revent);

enum
{ 
   GFX_K_PRT=1, 
   GFX_K_BACKSPACE,     GFX_K_TAB,      GFX_K_RETURN, GFX_K_ESCAPE,
   GFX_K_INSERT,        GFX_K_DELETE,   GFX_K_HOME,   GFX_K_END,
   GFX_K_PAGEUP,        GFX_K_PAGEDOWN,
   GFX_K_LEFT,          GFX_K_RIGHT,    GFX_K_DOWN,   GFX_K_UP,
   GFX_K_CONTROL=0x100, GFX_K_SHIFT= 0x200 , GFX_K_ALT= 0x400,
};

int gfx_init(char *name, int width, int height, char *title);
void gfx_close();


void gfx_geometry(unsigned int *W, unsigned int *H);
void gfx_set_sysbar_height(int hh);
void gfx_pause() ;
void gfx_resume() ;
void gfx_monitor_file(int fd, int ident, int R,int W,int X);
void gfx_stop_monitor_file(int fd,int ident);
void gfx_leave_screen();
void gfx_set_state(int wno, int state);
int  gfx_get_state(int wno);
int  gfx_get_selected(int wno);
void gfx_set_selected(int wno,int sel);
void gfx_set_text(int wno,char *str);
void gfx_set_fgbg(int wno, int fg, int bg);
void gfx_insert(int wno,int pos, char *text,void *user,unsigned int* icon);
void gfx_remove(int wno, int pos, void **Ruser);
char* gfx_get_text(int wno);
void gfx_clear(int wno);
void gfx_focus(int wno);
void gfx_text_data(int font, const char *str, 
                   int *Rw, int *Rh,int *Ra,unsigned char **Rd);
void gfx_scrollbar_set(int wno, double start, double end, double step);
void gfx_pixgrid_attribs(int pg, int *Rshown, unsigned int *Rw, unsigned int *Rh);
void gfx_pixgrid_put(int pg, int x,int y,
                     unsigned int *img, unsigned int iW,unsigned int iH,
                     int src_x, int src_y, 
                     int copy_w, int copy_h);
void gfx_pixgrid_putsimple(int pg, unsigned int *img);
 
int gfx_ask(char *title, char *butstr, char *fmt, ...);
char* gfx_select_file(int mode,char *init,char **types);
void gfx_canvas_setsize(int wno, int new_w, int new_h);
void gfx_canvas_draw_rectangle(int wno,int color, int x,int y,int w,int h);
void gfx_canvas_fill_rectangle(int wno,int color, int x,int y,int w,int h);
void gfx_canvas_recsel_begin(int wno, gfx_event_t *ev);
void gfx_canvas_recsel_end(int wno,int *Rx1, int *Ry1, int *Rx2, int *Ry2);
void gfx_canvas_print(int wno, int font, int color, int x, int y, char *str);
void gfx_canvas_put_graymap
       (int wno, unsigned int color, int dx,int dy, int w, int h,
        unsigned char *source, int sw, int sh, int sx, int sy);
void gfx_canvas_put
       (int wno, int dx,int dy, int w, int h,
        unsigned int *source, int sw, int sh, int sx, int sy);
int gfx_canvas_get_buffer
   (int wno, unsigned int *W, unsigned int *H, unsigned int **B);
void gfx_canvas_update(int wno, int x, int y, int w, int h);
void gfx_canvas_begin_report_mouse_position(int wno);
void gfx_canvas_end_report_mouse_position(int wno);
void gfx_canvas_set_zoom_params
  (int wno, int zf, int grid, unsigned int gridcolor,int w,int h);
void gfx_canvas_zoom(int wno,int x,int y);
void gfx_canvas_hide_zoom(int wno);
int gfx_sprite_new(int wno,int func, int w,int h, uint32_t *data);
void gfx_sprite_move(int wno,int sp, int x,int y);
void gfx_sprite_show(int wno,int sp,int sho);
void gfx_sprite_remove(int wno, int sp);
void gfx_spinner_set_imin(int wno, int min);
void gfx_spinner_set_imax(int wno, int max);
void gfx_spinner_set_iquant(int wno, int qua);
void gfx_spinner_set_fmin(int wno, double min);
void gfx_spinner_set_fmax(int wno, double max);
void gfx_spinner_set_fquant(int wno, double qua);
void gfx_spinner_set_fvalue(int wno, double value);
void gfx_spinner_set_ivalue(int wno, int value);
int gfx_spinner_get_ivalue(int wno);
double gfx_spinner_get_fvalue(int wno);
void gfx_logview_printf(int wno, const char *fmt, ...);
void gfx_logview_write(int wno, const char *txt, size_t len);
void gfx_logview_clear(int wno);
void gfx_set_log(int wno);
void loprintf(const char *fmt, ...);
void gfx_slider_set(int wno, double pos);
double gfx_slider_get(int wno);
void gfx_progressbar_total(int wno, int total);
void gfx_progressbar_set(int wno, int value);
void gfx_iconlist_insert
   (int wno, int pos,
    char *text, 
    unsigned int w, unsigned int h, unsigned int *data,
    int ival, void *pval);
void gfx_iconlist_updates(int wno, int upd);
void gfx_iconlist_area(int wno, int *Rw, int *Rh);
void gfx_iconlist_select(int wno, int item, int sel);
void gfx_iconlist_permute(int wno, int N, int *P);
void gfx_iconlist_clear_selection(int wno);
void gfx_iconlist_get_selection(int wno,int *Rn, int **Rs);
void gfx_filelist_set_directory(int wno,char *dir);
void gfx_filelist_set_types(int wno, char **types);
char** gfx_filelist_get_files(int wno);
char *gfx_filelist_get_directory(int wno);
void gfx_filelist_refresh(int wno);
void gfx_timer_start(int ident, int ms);
void gfx_timer_stop(int ident);

#ifdef GFX_ANDROID
#ifdef JNI_H_
JavaVM *gfx_get_javavm();
#endif
#ifdef ANDROID_NATIVE_ACTIVITY_H
ANativeActivity* gfx_get_activity();
#endif
void gfx_save_state(uint32_t len, uint8_t *data);
extern void (*gfx_app_stop)();
#endif

enum { _XX_WIDGETS_START=GFXEV_LAST+GFXRESERVED,
  /* WIDGET NAMES BELOW */
  Wcanvas,
  Wpen2,
  Wpen3,
  Wpen4,
  Wpen1,
  Wfile,
  Wprev,
  Wnext,
  Wadd,
  Wdel,
  Wset,
  Wpage,
  Wcut,
  Wins,
  WftTITLE,
  WftCLOSE,
  WftNEW,
  WftLIST,
  WptTITLE,
  WptGO,
  WptCLOSE,
  WptLIST,
  WptARRANGE,
  WamTITLE,
  WamDELALL,
  WamCLOSE,
  WamTEST,
  WamHELP,
  WpsTITLE,
  WpsOK,
  WpsCANCEL,
  WpsCANVAS,
  WpsCON,
  WpsSEL,
  WpsALL,
  WpsBLUE,
  WpsGREEN,
  WpsRED,
  WpsLAB2,
  WpsLAB1,
  WpsLAB0,
  WpsPE,
  WpsLAB3,
  WpsLAB4,
  WpsLAB5,
  WpsWIDTH,
  WpsEWIDTH,
  WpsEHEIGHT,
  WpsPAT,
  WsoERASE,
  WsoCUT,
  WsoCOPY,
  WsoEXPORT,
  WsoCANCEL,
  WioIMAGE,
  WioCLIPBOARD,
  WioCANCEL,
  WioTITLE,
  WcbTITLE,
  WcbCLOSE,
  WcbLIST,
  WprTITLE,
  WprPROGRESS,
  WseTITLE,
  WseEXPORT,
  WseDELFILE,
  WseCLOSE,
  WseCREDITS,
  WseRATE,
  WseSTYLE,
  WpoTITLE,
  WpoOK,
  WpoCANCEL,
  WpoLIST,
  WpoLAB0,

 _XX_WIDGETS_END };
void gfx_SCREEN_0();
void gfx_FILETHUMBS();
void gfx_PAGETHUMBS();
void gfx_ADMIN();
void gfx_PENSET();
void gfx_SELCLIPOP();
void gfx_SELINSOP();
void gfx_CLIPS();
void gfx_EXPORTPROG();
void gfx_SETTINGS();
void gfx_PAGEORDER();
extern unsigned int GRright[];
extern unsigned int GRleft2[];
extern unsigned int GRleft[];
extern unsigned int GRdel[];
extern unsigned int GRadd[];
extern unsigned int GRpen2[];
extern unsigned int GRfile[];
extern unsigned int GRclip[];
extern unsigned int GRcut[];
extern unsigned int GRcog[];
extern unsigned int GReras[];
extern unsigned int GRpen0[];
extern unsigned int GRpen1[];
extern unsigned int res0[];
