#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdint.h>
#include "gfx.h"
#include "shape.h"
#include "canvas.h"
#include "pngio.h"
#include "book.h"
#include "thumb.h"
#include "config.h"
#include "enviro.h"
#include "resh.h"
#include "rjpg.h"
#include "opdf.h"
#include "dbg.h"
#include "mkabout.h"
#include "shufle.h"

static int scr_w,scr_h;

static char bseq[8];

#define TOPBITS  0xFF000000

static int pageno=-1;
static unsigned int pagebg;
static char *current_file;
cfg_config_t config;

static book_t *mainfile;




static void show_page(int pn)
{
  unsigned int iw,ih,*id, cw,ch,*cd, dum;
  canvas_get_data(&cw, &ch, &cd);
  // pagebg= 0xFF000000;

  iw= cw; ih= ch; id= cd;
  if (book_read_image(mainfile, pn, 0, &iw, &ih, &id, &pagebg))
    canvas_clear();
  else if (iw>cw || ih>ch)
    canvas_set_data(iw, ih, id, pagebg);
  else
    canvas_update_data(iw, ih, pagebg);
}


void generate_white_noise(int w, int h, unsigned int *data)
{
  int x,y,dx,dy,T,tw,P,W;
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
   {
     data[x+y*w]= rand()%32;
   }
  static int matrix[]=
  {
    1, 1, 1,
    1, 1, 1,
    1, 1, 1,
  };
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
   {
     T= 0; tw= 0;
     for(dx=-1;dx<2;dx++)
       for(dy=-1;dy<2;dy++)
       {
         if ((x+dx<0) || (y+dy<0) || (x+dx>=w) || (y+dy>=w))
           P= 0;
         else P= data[x+y*w] & 0xff;
         W= matrix[(dx+1)+(dy+1)*3];
         T += P*W;
         tw+= W;
       }
      T /= tw;
      if (T<0) T=0; if (T>255) T= 255;
      data[x+y*w] |= (T<<16);
   }
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
   {
     P= (data[x+y*w] >> 16)&0xff;
     data[x+y*w]= 0xFF000000 | (P<<16) | (P<<8) | (P);
   }
}


static void admin_scr()
{
  gfx_event_t event;
  char *r;
  gfx_ADMIN();
  
again:
  if (!gfx_event(NULL,&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT:
  case WamCLOSE: goto done;
  case WamDELALL:
    del_all(storage_dir);
    break;
  case WamTEST:
    gfx_leave_screen();
{ 
  unsigned int *data,w,h;
  w= h= 300;
  data= malloc(w*h*sizeof(data[0]));
  generate_white_noise(w,h,data);
  canvas_insert_image(w,h,data);
}
    return ;
  }
  goto again;
done:
  gfx_leave_screen();
}

static void copy_page()
{
  int pn;
  pn= pageno-1;
  if (!mainfile || pn<0 || pn>=mainfile->npages) return ;
  if (book_copy_page(mainfile, pn, pageno)) return ;
  show_page(pageno);
  canvas_set_modified();
}


static void config_defaults()
{
#define CDEF(F,X) if (!config.G_##F) { config.G_##F= 1; config.F= X; }
  CDEF(bg,0)
  CDEF(thumb_w,128)
  CDEF(thumb_h,170)
#undef CDEF
#define PDEF(P,F,X) if (!config.P.G_##F) { config.P.G_##F= 1; config.P.F= X; }
  PDEF(ppen, color, 0xccFFff)
  PDEF(ppen, width, 2)
  PDEF(ppen, ewidth, 10)
  PDEF(ppen, eheight, 40)
  
  PDEF(philight, color, 0xFF0033)
  PDEF(philight, width, 2)
  PDEF(philight, ewidth, 10)
  PDEF(philight, eheight, 40)
  
  PDEF(ptitle, color, 0x3300ff)
  PDEF(ptitle, width, 10)
  PDEF(ptitle, ewidth, 10)
  PDEF(ptitle, eheight, 40)
#undef PDEF
}

void save_config()
{
  char buf[300];
  if (storage_dir) 
  {
    sprintf(buf,"%s/config.txt",storage_dir);
    cfg_write_config(buf, &config);
  }
}

static void load_config()
{
  char buf[300];
  if (storage_dir) 
  {
    sprintf(buf,"%s/config.txt",storage_dir);
    cfg_parse_config(buf, &config);
  }
  config_defaults();
}





static int seqsfx(char *k)
{
  int z,i;
  z= strlen(k);
  for(i=0;i<z;i++)
    { if (k[z-1-i]!=bseq[6-i]) return 0; }
  return 1;
}

static int setstyle(int elt, int ret)
{
  int SS;
  SS= style_setting(elt, &pagebg);
  if (SS&1)
  {
    uint32_t caw, cah, *cad;
    canvas_get_data(&caw, &cah, &cad);
    canvas_update_data(caw, cah,pagebg);
    if ((SS!=3) && mainfile)
      book_set_bg(mainfile, pageno, pagebg);
  }
  if ((SS&2) && mainfile)
    book_set_bgall(mainfile, pagebg);
  return ret;
}

int SEQ(int chr)
{
  int i;
  for(i=0;i<6;i++) bseq[i]= bseq[i+1];
  bseq[6]= chr;
  if (seqsfx("TP>TPR<")) { admin_scr(); return 1; }
  if (seqsfx("++")) { copy_page(); return 1; }
  if (seqsfx("EE")) 
  {
     if (gfx_ask("Why you double tap?", "!OK!Cancel",
               "Whole page will be cleared!")==0)
       { canvas_clear(); canvas_set_modified(); }
     return 1; 
  }
  if (seqsfx("TT"))  return setstyle(0, 1); 
  if (seqsfx("PP"))  return setstyle(1, 1);
  if (seqsfx("RR"))  return setstyle(2, 1); 
  return 0;
}

static void save_page()
{
  uint32_t *pdata, pw,ph;
  uint32_t *tdata, tw,th;

  if (pageno<0) return ;
  if (!canvas_modified() || !mainfile) return ;
  canvas_get_data(&pw,&ph,&pdata);
  book_update_image(mainfile, pageno, 0, pw, ph, pdata, pagebg);

  canvas_get_fdata(&pw,&ph,&pdata);
  if (resize_image_argb(pdata, pw, ph, 
                        config.thumb_w, config.thumb_h,
                        &tdata, &tw, &th))
     return ;
  book_update_image(mainfile, pageno, 1, tw, th, tdata, pagebg);
  free(tdata);
}

static void save_all()
{
  DPRINT(("saveall: save_page\n"))
  save_page();
}

static void set_page(int pn)
{
  char buf[20];
  if (!mainfile || pn<0 || pn>=mainfile->npages || pn==pageno) return ;
  save_page();
  show_page(pn);
  pageno= pn;
  sprintf(buf,"%d/%d", pageno+1,mainfile->npages);
  gfx_set_text(Wpage,buf);
}

int create_note_file(char *fn)
{
  book_t *BO;
  if (book_create(fn, &BO))
  {
    gfx_ask("Error", "!OK", "can't create %s\n", fn);
    return 1;
  }
  book_close(BO);
  return 0;
}

static void close_note_file()
{
  if (!mainfile) return ;
  save_page();
  book_cleanup(mainfile);
  book_close(mainfile);
  mainfile= NULL;
  if (current_file) free(current_file);
  current_file= NULL;
}

static int load_note_file(char *fn,int pn,int ro) 
{
  book_t *BB;
  if (current_file && !strcmp(fn,current_file)) return 0;
  
  if (book_read(fn, &BB)) return 1;

  close_note_file();
  mainfile= BB;
  BB->readonly= ro;
  current_file= strdup(fn);

  if (pn<0) pn= 0; 
  if (pn>=mainfile->npages) pn= mainfile->npages-1;
  pageno= -1;
  set_page(pn);  
}

static void add_page()
{
  if (!mainfile) return ;
  if (book_make_page(mainfile, pageno+1)) return ;
  set_page(pageno+1);
}

static void remove_page()
{
  int i,pn;
  if (!mainfile || mainfile->npages==1) return ;
  if (gfx_ask("Confirm Page Deletion","!Yes!No",
              "Do you really want to delete this page?")) return ;
  if (book_del_page(mainfile,pageno)) return ;
  
  pn= pageno;
  if (pn>=mainfile->npages) pn--;
  pageno= -1;
  set_page(pn);
}

static unsigned int *ialloc(int w,int h)
{ 
  return malloc(sizeof(unsigned int)*w*h);
}

static void populate_page_list()
{
  int i;
  unsigned int *dd, *d;
  unsigned int iw, ih, *id;
  char buf[20];

  dd= NULL;
  iw= config.thumb_w; ih= config.thumb_h;
  id= ialloc(iw,ih); 
  gfx_iconlist_updates(WptLIST,0);

  for(i=0;i<mainfile->npages;i++)
  {
    if (make_page_thumbnail(mainfile,i, iw, ih, id))
    {
      if (!dd) { dd= ialloc(iw,ih); make_default_thumbnail(iw,ih,dd); }
      d= dd;
    }
    else d= id;
    sprintf(buf,"%03d", i+1);
    gfx_iconlist_insert(WptLIST,-1,buf,iw,ih,d,i,NULL);
  }

  if (dd) free(dd);
  free(id);
  gfx_iconlist_updates(WptLIST,1);
}

static void page_screen()
{
  gfx_event_t event;
  int pn;
  int NP,*P;

  pn= -1;
  if (!mainfile) return ;

  gfx_PAGETHUMBS();
  populate_page_list();

again:
  if (!gfx_event(NULL,&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT:
  case WptCLOSE: if (pageno!=-1) pn= -1; goto done;
  case WptLIST: pn= event.key; break;
  case WptGO: if (pn!=-1) goto done; else break;
  case WptARRANGE: 
    if (reorder_pages(mainfile,&NP,&P)) break;
    if (!book_permute(mainfile, NP, P))
    {
      int i;
      for(i=0;i<NP;i++) if (P[i]==pageno) { pageno= i; break; }
      gfx_clear(WptLIST);
      populate_page_list();
      pn= pageno;
      pageno= -1;
    }
    free(P);
    break;
  }
  goto again;
done:
  gfx_leave_screen();
  if (pn!=-1) set_page(pn);
}

#ifdef GFX_ANDROID

static void state(char *buf)
{
  if (current_file)
    sprintf(buf,"%d.%s",pageno, current_file);
  else
    sprintf(buf,"-1");
}

static void save_state()
{
  char buf[2000];
  state(buf);
  DPRINT(("state: [%s]\n",buf))
  gfx_save_state(strlen(buf)+1, (uint8_t*) strdup(buf));
}

static void app_stop()
{
  char buf[2000];
  char ff[500];
  FILE* f;

  DPRINT(("stopping\n"))
  save_page();
  state(buf);
  if (storage_dir)
  {
    sprintf(ff,"%s/last.txt",storage_dir);
    f= fopen(ff,"wb"); if (f) { fprintf(f,"%s\n",buf); fclose(f); }
  }
  if (mainfile)
    book_cleanup(mainfile);
}
#endif

static int ask_insert_source()
{
  gfx_event_t event;
  int R;
  gfx_SELINSOP();
  R= 0;
  while(!R)
  {
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
     case WioIMAGE: case WioCLIPBOARD: case WioCANCEL:
       R= event.type;
       break;
    }
  }
  gfx_leave_screen();
  return R;
}

int sys_wait_event(gfx_event_t *event)
{
#ifdef GFX_ANDROID
  switch(event->type)
  {
   case GFXEV_QUIT: return 1;
   case GFXEV_SAVE_STATE: save_state(); break;
  }
#endif
  return 0;
}

static int has_suffix(char *str, char *sfx)
{
  int Ls, Lf;
  Ls= strlen(str);
  Lf= strlen(sfx);
  if (Ls<=Lf) return 0;
  str+= Ls - Lf;
  return strcasecmp(str,sfx)==0;
}

static void insert_image()
{
  char *fn;
  unsigned int w,h,*d;
#ifdef GFX_ANDROID
  fn= call_sys_str("choose_image_file", 1, "cif_s1");
#endif
#ifdef GFX_X11
  static char *ftypes[]=
  {
    "Image Files",
    " JPEG Images (.jpg .jpeg)",
    " Portable Network Graphics (.png)",
    NULL
  };
  fn= gfx_select_file(0, NULL, ftypes);
#endif
  if (!fn) return ;
  d= NULL;
  if (has_suffix(fn,".png"))
  {
    char *e;
    e= read_png_image(fn,&w,&h,&d);
    if (e)
    {
      gfx_ask("Error", "!OK", "Can't read PNG image: %s", e);
      return ;
    }
  }
  else if (has_suffix(fn,".jpg") || has_suffix(fn,".jpeg"))
  {
    int i;
    if (read_jpeg(fn, (int*) &w, (int*)  &h, (int**) &d))
    {
      gfx_ask("Error", "!OK", "Can't read JPEG image");
      return ;
    }
    for(i=0;i<w*h;i++) d[i]|= 0xFF000000;
  }
  else
  {
      gfx_ask("Error", "!OK", "Can't load image: unknown format");
      return ;
  }
  canvas_insert_image(w,h,d);
}

static char *export_fn()
{
  char buf[300];
  char *z,*k;
  if (!current_file || !export_dir) return NULL;
  z= current_file;
  while ((k=strchr(z,'/'))) z=k+1;
  z= strdup(z);
  k= strchr(z,'.');
  if (k) k[0]= 0;
  sprintf(buf,"%s/export-%s.pdf",  export_dir, z);
  free(z); 
  return strdup(buf);
}

static void image_fit
   (int iw, int ih, int pw, int ph, int *Rsw, int *Rsh, int *Rx, int *Ry)
{
  if ( iw/(double)ih > pw / (double) ph)
  {
    *Rsw= pw;
    *Rsh= (double) ih / iw * pw;
  }
  else
  {
    *Rsh= ph;
    *Rsw= (double) iw / ih * ph;
  }
  *Rx= (pw-*Rsw)/2;
  *Ry= (ph-*Rsh)/2;
}

static void apply_bg
  (unsigned int bg, unsigned int *D, unsigned int W,unsigned int H)
{
  unsigned int *bD, bW, bH, bM;
  unsigned int B, bR, bG, bB;
  unsigned int P, pR, pG, pB, pA;
  unsigned int x,y,k;

  gen_pattern(bg,&bW,&bH,&bD,&bM);
  for(y=0;y<H;y++)
    for(x=0;x<W;x++)
    {
      B= bD[ (x%bW) + (y%bH)*bW ];
      k= x+y*W;
      P= D[k];
      pA= P>>24;
      if (pA==0) { D[k]= B; continue; }
      if (pA==255) { D[k]= P; continue; }
      pR= (P>>16)&0xff; pG= (P>>8)&0xff; pB= (P)&0xff; 
      bR= (B>>16)&0xff; bG= (B>>8)&0xff; bB= (B)&0xff; 
      pR= (pR*pA+bR*(255-pA))/255;
      pG= (pG*pA+bG*(255-pA))/255;
      pB= (pB*pA+bB*(255-pA))/255;
      D[k]= (pR<<16) | (pG<<8) | (pB);
    }
}


static void export_file()
{
  char *fn, *z;
  int i,j, sw, sh, sx, sy;
  unsigned int iw,ih,*idata;
  char *decode_err,*encode_err;
  int pw, ph;
  unsigned int elen; unsigned char *edata;
  int image;
  unsigned int ibg;

  if (!mainfile || !(fn=export_fn())) return ;
  save_all();
  if (opdf_begin(fn)) return ;
  gfx_EXPORTPROG();
  pw= 595; ph= 842;
  gfx_progressbar_total(WprPROGRESS, mainfile->npages);
  for(i=0;i<mainfile->npages;i++)
  {
     iw= ih= 0; 
     if (book_read_image(mainfile,i,0,&iw,&ih,&idata,&ibg)) continue;
     apply_bg(ibg, idata, iw, ih);
     encode_err= encode_png_rgb(iw, ih, idata, &elen, &edata);
     if (encode_err) continue;
     opdf_begin_page(pw,ph);
     opdf_pngmem(edata, elen, &image, NULL, NULL);
     image_fit(iw, ih, pw, ph, &sw, &sh, &sx, &sy);
     opdf_draw_image(image, sx, sy, sw, sh);
     opdf_end_page();
     free(edata);
     gfx_progressbar_set(WprPROGRESS, i+1);
  }
  opdf_end();
  gfx_leave_screen();
  while((z=strchr(fn,'/'))) fn= z+1;
  canvas_print_tempo(3500, "Exported to %s in Downloads folder", fn);
}

static void settings_screen()
{
  gfx_event_t event;
  uint32_t w,h, *d;
  char *ff;

  gfx_SETTINGS();
again:
  if (!gfx_event(NULL,&event)) goto again;
  switch(event.type)
  {
  case WseEXPORT: break;
  case WseCLOSE: break;
  case WseCREDITS: break;
  case WseDELFILE: break;
  case WseSTYLE: break;
  case WseRATE: break;
  default: goto again;
  }
  gfx_leave_screen();
  switch(event.type)
  {
  case WseEXPORT: export_file(); break;
  case WseCREDITS: 
    save_all();
    if (current_file && !strcmp(current_file,"about.not")) break;
    canvas_get_data(&w,&h,&d); 
    make_about_file(w,h);
    load_note_file("about.not",0, 1);
    break;
  case WseDELFILE: 
    if (!mainfile || !current_file) break;
    if (gfx_ask("Confirm Deletion", "!Yes!No", 
          "Do you really want to delete this file?")) break;
    ff= strdup(current_file);
    close_note_file(mainfile);
    book_delete(ff);
    while(!(ff= choose_file(config.thumb_w, config.thumb_h, 
                       storage_dir, NULL)));
    load_note_file(ff,0,0);
    break;
  case WseRATE: 
#ifdef GFX_ANDROID
    //canvas_print_tempo(3500,"Many thanks!"); 
    call_sys_void("rateApp");
#endif
    break;
  case WseSTYLE: setstyle(0,0); break;
  }
}

static void main_loop()
{
  char *fn;
  gfx_event_t event;
  select_pen(1);
  while(1)
  {
    if (!gfx_event(NULL,&event)) continue;
    if (canvas_event(&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT: 
       DPRINT(("Quitting\n"))
       return ;
    case Wadd: if (SEQ('+')) break;
               add_page(); break;
    case Wdel: remove_page(); break;
    case Wprev: if (SEQ('<')) break; set_page(pageno-1); break;
    case Wnext: if (SEQ('>')) break; set_page(pageno+1); break;
    case Wfile: 
       fn= choose_file(config.thumb_w, config.thumb_h, 
                       storage_dir, current_file);
       if (!fn) break;
       load_note_file(fn,0,0);
       break;
    case Wpage: page_screen(); break;
    case Wset:  settings_screen(); break;
    case Wcut:  canvas_begin_select(); break;
    case Wins:
       switch(ask_insert_source())
       {
         case WioIMAGE: insert_image(); break;
         case WioCLIPBOARD: insert_from_clipboard(); break;
         case WioCANCEL: break;
       }
       break;
#ifdef GFX_ANDROID
    case GFXEV_SAVE_STATE:
       DPRINT(("save state now!\n"))
       save_state();
       break;
#endif
    case GFXEV(500,TIMEOUT):
       // DPRINT(("timeout!!\n"))
   { static int v; char buf[200]; sprintf(buf,"%c", v?'x':'X');
    gfx_set_text(Wins, buf); v= ~v;
       gfx_timer_start(500, 1500); }
       break;
    }
  }
}

void export_image(int w, int h, unsigned int *d,unsigned int bg)
{
  char fname[400];
  int fno;
  char *e;
  int i;
  if (!export_dir) 
  {
    gfx_ask("Error", "!OK", "Internal error: no export dir");
    return ;
  }
  for(fno=1;fno<200;fno++)
  {
    sprintf(fname,"%s/note-export%03d.png", export_dir, fno);
    if (access(fname,F_OK) && errno==ENOENT) break;
  }
  if (fno==200) 
  {
    gfx_ask("Error", "!OK", "Internal error: can't access directory");
    return ;
  }
  apply_bg(bg, d, w,h);
  for(i=0;i<w*h;i++) d[i]|=0xFF000000;
  e= write_png_image(fname,w,h,d);
  if (e)
    gfx_ask("Error", "!OK", "PNG write: %s",e);
  else
    canvas_print_tempo(4500, "Exported image data to "
                           "note-export%03d.png"
                           " in Downloads folder", fno);
}


#ifdef GFX_ANDROID

void android_main(void *state, size_t state_size)
{
  char *fn;
  char sbuf[2000];
  int pn;
  config.thumb_w= 128; config.thumb_h= 170;
  init_enviro();
#ifdef DBG
  dbg_init();
#endif
  load_config();
  gfx_app_stop= app_stop;
  gfx_SCREEN_0();
  canvas_init(&scr_w,&scr_h);

  pn= -1;
again:
  if (state) { char *e; pn= strtol(state,&e,0);
               if (pn<0) goto choose;
               if (e[0]!='.') goto choose;
               fn= e+1; goto doit; }
  if (storage_dir)
  {
   char ff[500];
   sprintf(ff,"%s/last.txt",storage_dir);
   FILE* f;
   f= fopen(ff,"rb"); 
   if (f) {
     fgets(sbuf,sizeof(sbuf),f);
     int i; for(i=0;sbuf[i];i++)
        if (sbuf[i]=='\r' || sbuf[i]=='\n' || i==sizeof(sbuf)-1)
           { sbuf[i]=0; break; } 
      fclose(f); 
      state= sbuf; 
      goto again;
   }
  }

choose:
  fn= choose_file(config.thumb_w, config.thumb_h, 
                  storage_dir, current_file);
  if (!fn) goto finish;

doit:
  DPRINT(("File %s, page %d\n", fn, pn))
  load_note_file(fn,pn,0);
  if (pn!=-1) set_page(pn);
  main_loop();

finish:
  return ;
}

#endif

#ifdef GFX_X11

int main(int argc,char **argv)
{
  gfx_init(NULL,768,900,"Notes");
  gfx_SCREEN_0();
  init_enviro();
#ifdef DBG
  dbg_init();
#endif
  load_config();
  canvas_init(&scr_w,&scr_h);

  if (argc>1)
  {
    if (!strcmp(argv[1],"-c"))
    {
      create_note_file(argv[2]);
      goto finish;
    }
    load_note_file(argv[1],0,0);
  }

  main_loop();

  if (mainfile) 
  {
    save_page();
    book_cleanup(mainfile);
    book_close(mainfile);
  }
finish:
  gfx_close();
  return 0;
}
#endif
