#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "shape.h"
#include "gfx.h"
#include "config.h"
#include "resh.h"
#include "genpat.h"

#define PATPRESIZ  80
#define ERASIZ     80.0
#define PENMAXSIZ  50.0



static void render_resh
  (int wi, int fg, int bg,
   unsigned int **Rdata, unsigned int *Rw, unsigned int *Rh);


static void resh_init(int widgetno, int bg);
static void resh_update_bg(int bg);
static void resh_update_stroke(int pno, int wi, int co);
static void resh_update_eraser(int pno, int wi, int he);



typedef struct { int code, x,y; } resh_t;
typedef struct 
{
  resh_t *points; 
  int x,y,w, h; 
  int ax, ay, aw, ah;
} sec_t;

static int widget;
static unsigned int caw, cah, *cad;
static unsigned int dum;
static unsigned int bw=1, bh=1, *bd= &dum, bmark= 0xFFffFF;
static cfg_pen_t pens[3];


#include "setimg.ci"

static sec_t  secs[3]= 
{
  { REtitle },
  { REnormal },
  { REred }
};

static unsigned int *data, w,h;

static void pixel(int x,int y, unsigned char value, int color)
{
  unsigned int P, k;
  unsigned int pR,pG,pB, cR,cG,cB;

  if (x<0 || y<0 || x>=caw || y>=cah) return ;
  k= x+y*caw;

  if (value==255) { cad[k]= color; }
  else if (value!=0)
  {
     P= cad[k];
     pR= P>>16;     pG= P>>8;     pB= P;
     pR &= 0xff;    pG &= 0xff;   pB &= 0xff;
     cR= color>>16; cG= color>>8; cB= color;
     cR &= 0xff;    cG &= 0xff;   cB &= 0xff;

     pR= (cR*value+pR*(255-value)) / 255;
     pG= (cG*value+pG*(255-value)) / 255;
     pB= (cB*value+pB*(255-value)) / 255;

     cad[k]= (pR<<16) | (pG<<8) | (pB);
   }
}

#include "brushline.c"
#include "aathick.c"

static double thickness(double *K, int d1, int d2) { return *K; }

static void drawAAline(int x0,int y0,int x1,int y1,int color)
{
  double T;
  T= 3.2/2;
  draw_varthick_line(x0,y0,x1,y1, (void*) thickness, &T,
                                  (void*) thickness, &T, color);
}

static resh_t timg[]= {
{'d',  91, 130},
{'m',  196, 234},
{'m',  195, 234},
{'m',  193, 234},
{'m',  193, 233},
{'e',  0, 0 },
};

static int imin2(int a,int b) { return a<b?a:b; }
static int imax2(int a,int b) { return a>b?a:b; }

static void normalize_input(sec_t *S)
{
  resh_t *R;
  int i;
  int mix,max,miy,may;
  if (S->w) return ;
  R= S->points;
  max= mix= R[0].x;
  may= miy= R[0].y;
  for(i=0;R[i].code!='e';i++)
  {
    max= imax2(max, R[i].x);
    may= imax2(may, R[i].y);
    mix= imin2(mix, R[i].x);
    miy= imin2(miy, R[i].y);
  }

  for(i=0;R[i].code!='e';i++)
  {
    R[i].x -= mix;
    R[i].y -= miy;
  }
  S->w= max-mix+1;
  S->h= may-miy+1;
}

static void layout_sections()
{
  int i,j, y,tw,th;
  resh_t *R;
  y= 0;
  tw= 0;
  if (secs[0].ah) return ;
  for(i=0;i<3;i++)
  {
    normalize_input(secs+i);
    secs[i].ay= y;
    secs[i].y= secs[i].ay+30;
    y= secs[i].y + secs[i].h + 30;
    secs[i].ah= secs[i].h + 60;
    secs[i].x= 30;
    secs[i].ax= 0;
    secs[i].aw= secs[i].w + 60;
                  // +40 is for the eraser 
    tw= imax2(secs[i].aw + 40, tw);
  }
  th= y;

  for(i=0;i<3;i++)
  {
    secs[i].ax+= (caw-tw)/2;
    secs[i].x+= (caw-tw)/2;
    secs[i].ay+= (cah-th)/2;
    secs[i].y+= (cah-th)/2;
    R= secs[i].points;
    for(j=0;R[j].code!='e';j++)
      { R[j].x+= secs[i].x; R[j].y+= secs[i].y; }
  }
}

static void paintbg(int x,int y,int w, int h)
{
  int r,c,xx,yy;
  if (x<0) { w+= x; x= 0; }
  if (y<0) { h+= y; y= 0; }
  if (x+w>caw) w= caw-x;
  if (y+h>cah) h= cah-h;

  for(r=0;r<h;r++)
   for(c=0;c<w;c++)
   {
      yy= r+y;
      xx= c+x;
      cad[ xx+ yy*caw ]= bd[ xx%bw + (yy%bh)*bw ];
   }
}

static void render_strokes(sec_t *S,int upd, int color, int width)
{
  int x,y;
  resh_t *R;
  shape_t *sha;
  int initpo, i;

  R= S->points;
  if (upd)
    paintbg(S->ax, S->ay, S->aw, S->ah);

  if (width<=3) sha= NULL;
           else sha= pen_new(width);
  
  for(i=0;R[i].code!='e';i++)
  {
    if (R[i].code=='m')
    {
      if (sha) brush_line(initpo, sha, x, y, R[i].x, R[i].y, color);
          else drawAAline(x,y, R[i].x, R[i].y, color);
      initpo= 1;
    } else initpo= 0;
    x= R[i].x;
    y= R[i].y;
  }

  if (sha) shape_free(sha);
  if (upd)
    gfx_canvas_update(widget, S->ax, S->ay, S->aw, S->ah);
}

static void render_eraser(sec_t *S, int upd, int wi, int he)
{
  int ex,ey,x,y;
  if (upd)
    paintbg(S->ax+S->aw, S->ay, caw-S->ax-S->aw, S->ah);
  for(ey=0;ey<he;ey++)
  for(ex=0;ex<wi;ex++)
  {
    x= ex+S->ax+S->aw;
    y= ey+S->ay;
    if (x<0 || y<0 || x>=caw || y>=caw) continue;
    cad[x+y*caw] ^= 0xFFffFF;
  }
  if (upd)
    gfx_canvas_update(widget,S->ax+S->aw, S->ay, caw-S->ax-S->aw, S->ah);
}

static void resh_init(int widgetno, int bg)
{
  widget= widgetno;
  gfx_canvas_get_buffer(widget, &caw, &cah, &cad);
  layout_sections();
  pens[0]= config.ptitle;
  pens[1]= config.ppen;
  pens[2]= config.philight;
  resh_update_bg(bg);
}

static unsigned int current_bg;

static void resh_update_bg(int bg)
{
  int i;
  current_bg= bg;
  gen_pattern(bg, &bw, &bh, &bd, &bmark);
  paintbg(0,0,caw, cah);
  for(i=0;i<3;i++)
  {
    render_strokes(secs+i, 0, pens[i].color, pens[i].width);
    render_eraser(secs+i, 0, pens[i].ewidth, pens[i].eheight);
  }
  gfx_canvas_update(widget, 0, 0, caw, cah);
}

static void resh_update_stroke(int pno, int wi, int co)
{
  render_strokes(secs+pno, 1, co, wi);
  pens[pno].width= wi;
  pens[pno].color= co;
}

static void resh_update_eraser(int pno, int wi, int he)
{
  render_eraser(secs+pno, 1, wi, he);
  pens[pno].ewidth= wi;
  pens[pno].eheight= he;
}

static void render_resh
  (int wi, int fg, int bg,
   unsigned int **Rdata, unsigned int *Rw, unsigned int *Rh)
{
  int i;
  int mix,max,miy,may,ox,oy,x,y;
  shape_t *S;
  miy= mix= 1024; 
  may= max= -1;

  S= pen_new(wi);
  for(i=0;timg[i].code!='e';i++)
  {
     miy= imin2(miy, timg[i].y);
     may= imax2(may, timg[i].y);
     mix= imin2(mix, timg[i].x);
     max= imax2(max, timg[i].x);
  }

  w= (max-mix)+2*wi+20;
  h= (may-miy)+2*wi+20;

  data= malloc(w*h*sizeof(data[0]));
  for(i=0;i<w*h;i++) data[i]= bg;

  ox= -mix + wi+1;
  oy= -miy + wi+1;

  for(i=0;;i++)
    switch(timg[i].code)
    {
     case 'e': goto done;
     case 'm': brush_line(0,S,x+ox,y+oy,timg[i].x+ox,timg[i].y+oy,fg);
     case 'd': x= timg[i].x; y= timg[i].y; break;
    }

done:
  *Rdata= data;
  *Rw= w;
  *Rh= h;
  shape_free(S);
}

#if 0
static double dmax2(double a, double b) { return a>b? a:b; }
static double dmin2(double a, double b) { return a<b? a:b; }
#endif

static double col2sli(unsigned int K,int Wi) 
   { gfx_slider_set(Wi, (K&0xff)/255.0); }

static unsigned int sli2col(int Wi)
   { return gfx_slider_get(Wi)*255.0; }

static void s_color(unsigned int val)
{
  col2sli(val>>16, WpsRED);
  col2sli(val>>8, WpsGREEN);
  col2sli(val, WpsBLUE);
}

static unsigned int g_color()
{
  return (sli2col(WpsRED)<<16) | (sli2col(WpsGREEN)<<8) | sli2col(WpsBLUE);
}

static int selsec;

static int patlist_ok;

static void populate_pattern_list()
{
  int i,n;
  // int aw, ah;
  unsigned int tw, th, *td;
  char *name;
  tw= PATPRESIZ; th= PATPRESIZ;
  td= malloc(sizeof(td[0])*tw*th);
  // gfx_iconlist_area(WpsPAT, &aw, &ah);
  
  gfx_iconlist_updates(WpsPAT, 0);
  n= get_npatterns();

  for(i=0;i<n;i++)
  {
    get_pattern_info(i,tw,th,td,&name);
    gfx_iconlist_insert(WpsPAT, -1, name, tw, th, td, 0, NULL);
  }
  gfx_iconlist_updates(WpsPAT, 1);
  free(td);
}

static void select_panels()
{
  switch(selsec)
  {
  case 0: 
  case 1:
  case 2:
    gfx_set_selected(WpsCON, 0);
    gfx_set_selected(WpsPE, 0);
    break;
  case 3:
    gfx_set_selected(WpsCON, 0);
    gfx_set_selected(WpsPE, 1);
    break;
  case 4:
    gfx_set_selected(WpsCON, 1);
    if (!patlist_ok) populate_pattern_list();
    patlist_ok= 1;
    break;
  }
}

typedef struct
{
  unsigned int *color;
  int color_action;
  unsigned int *width;
  unsigned int *ewidth;
  unsigned int *eheight;
} edit_t;

static unsigned int dummy, bgcolor, bgpat;
static cfg_config_t CC;

static edit_t edits[]=
{
 { &CC.ptitle.color, 'D', &CC.ptitle.width, &CC.ptitle.ewidth,
   &CC.ptitle.eheight },
 { &CC.ppen.color, 'D', &CC.ppen.width, &CC.ppen.ewidth,
   &CC.ppen.eheight },
 { &CC.philight.color, 'D', &CC.philight.width, &CC.philight.ewidth,
   &CC.philight.eheight },
 { &bgcolor, 'B', &dummy, &dummy, &dummy },
 { &dummy, 'I', &dummy, &dummy, &dummy }
};
static edit_t *EE;

static void select_section(int p)
{
  selsec= p;
  EE= edits+p;
  select_panels();
  gfx_set_selected(WpsSEL, p);
}



int style_setting(int p,unsigned int* bg)
{
  gfx_event_t event;
  unsigned int orig_bg;
  int R;

  orig_bg= *bg;

  CC= config;
  gfx_PENSET();

  gfx_insert(WpsSEL,-1,"Title Pen", NULL, NULL);
  gfx_insert(WpsSEL,-1,"Normal Pen", NULL, NULL);
  gfx_insert(WpsSEL,-1,"Red Pen", NULL, NULL);
  gfx_insert(WpsSEL,-1,"Page Color", NULL, NULL);
  gfx_insert(WpsSEL,-1,"Page Pattern", NULL, NULL);

  resh_init(WpsCANVAS, orig_bg);
  patlist_ok= 0;

  bgcolor= orig_bg;
  bgpat= orig_bg;
  R= 0;

restart:
  select_section(p);
  s_color(EE->color[0]);
  gfx_slider_set(WpsWIDTH, EE->width[0]/PENMAXSIZ);
  gfx_slider_set(WpsEWIDTH, (EE->ewidth[0]-1)/ERASIZ);
  gfx_slider_set(WpsEHEIGHT, (EE->eheight[0]-1)/ERASIZ);
  
again:
  if (!gfx_event(NULL,&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT:
  case WpsCANCEL:
    goto cancelled;
  case WpsRED:
  case WpsGREEN:
  case WpsBLUE:
    if (EE->color_action=='I') goto again;
    EE->color[0]= g_color();
    if (EE->color_action=='D')
      resh_update_stroke(selsec, EE->width[0], EE->color[0]);
    else
      resh_update_bg(EE->color[0]);
    goto again;
  case WpsPAT:
    resh_update_bg( event.key | 0x01000000 );
    goto again;
  case WpsWIDTH:
    EE->width[0]= PENMAXSIZ*gfx_slider_get(WpsWIDTH);
    resh_update_stroke(selsec, EE->width[0], EE->color[0]);
    goto again;
  case WpsEWIDTH:
    EE->ewidth[0]= ERASIZ*gfx_slider_get(WpsEWIDTH)+1;
    resh_update_eraser(selsec, EE->ewidth[0], EE->eheight[0]);
    goto again;
  case WpsEHEIGHT:
    EE->eheight[0]= ERASIZ*gfx_slider_get(WpsEHEIGHT)+1;
    resh_update_eraser(selsec, EE->ewidth[0], EE->eheight[0]);
    goto again;
  default:
    goto again;
  case WpsOK:
    config= CC;
    canvas_reload_config();
    save_config();
    break;
  case WpsSEL:
    p= event.key;
    goto restart;
  }
  if (orig_bg!=current_bg)
  {
    *bg= current_bg;
    R|= 1;
  }
  if (gfx_get_state(WpsALL)) R|= 2;
cancelled:
  gfx_leave_screen();
  return R;
}


