#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "shape.h"


static int dw,dh;
static unsigned char *data;
static int npoints;
static dpoint_t *points;

#ifdef SELF_TEST
static int mx,my,Mx,My;
static int imin2(int a,int b) { return a<b?a:b; }
static int imax2(int a,int b) { return a>b?a:b; }
#endif

#ifdef SELF_TEST
static void prt_data()
{
  int x,y;
  for(y=0;y<dh;y++)
  {
    for(x=0;x<dw;x++) if (data[x+y*dw]) printf("X");
                                  else printf(" ");
    printf("\n");
  }
}
#endif

static void pixel(int x,int y,unsigned char value)
{
  if (x<0 || y<0 || x>=dw || y>=dh) return ;
  data[x+y*dw]= value;
}

static void fillcircle(int r)
{
  int x,y,sy,sx,ky,kx,count;
  int R;
  R= r/2;
  for(y=-R;y<=R;y++)
   for(x=-R;x<=R;x++)
   {
     count= 0;
     for(sy=-2;sy<2;sy++)
      for(sx=-2;sx<2;sx++)
      {
        ky= 2*(4*y+sy)+1; kx= 2*(4*x+sx)+1;
        if ((kx*kx+ky*ky)<=(16*r*r))
         count++;
      }
     count= count * 255 / 16;
     pixel(x+R,y+R,count);
   }
}

static void make_stroke(stroke_t *R)
{
  int S;
  R->len= npoints;
  if (!R->len) { R->data= NULL; return ; }
  S= sizeof(R->data[0])*npoints;
  R->data= malloc(S);
  memcpy(R->data, points, S);
}

static void dpoint(int x,int y, unsigned char v)
{
  points[npoints].dx= x;
  points[npoints].dy= y;
  points[npoints].value= v;
  npoints++;
}

static void shape_diff(int dx,int dy, stroke_t *R,int cx, int cy)
{
  int x,y, ox,oy;
  unsigned char P,V;
  npoints= 0;
#if 1
  for(y=0;y<dh;y++)
    for(x=0;x<dw;x++)
    {
      P= data[x+y*dw];
      if (!P) continue;
      ox= x+dx; oy= y+dy;
      if (ox<0 || ox>=dw || oy<0 || oy>=dh)
        V= 0;
      else
        V= data[ox+oy*dw];
      if (V>=P) continue; // already filled in with enough color
      dpoint(x-cx,y-cy, P);
    }
#else
  for(y=0;y<dh;y++)
    for(x=0;x<dw;x++)
    {
      P= data[x+y*dw];
      dpoint(x-cx, y-cy, P);
    }
#endif
  
  make_stroke(R);
}

static void copy_data(stroke_t *S,int cx,int cy)
{
  int x,y;
  npoints= 0;
  for(y=0;y<dh;y++)
    for(x=0;x<dw;x++)
      if (data[x+y*dw]) dpoint(x-cx,y-cy,data[x+y*dw]);
  make_stroke(S);
}

static void alloc_data(int w,int h)
{
  dw= w;
  dh= h;
  data= malloc(dw*dh);
  points= malloc(dw*dh*sizeof(points[0]));
}

static shape_t* make_all_strokes(int cx,int cy)
{
  int dx,dy;
  shape_t *H;

  H= calloc(1,sizeof(*H));
  for(dx=-1;dx<2;dx++)
   for(dy=-1;dy<2;dy++)
    if (dx || dy) 
      shape_diff(dx, dy, H->strokes + (dy+1)*3+(dx+1), cx, cy);
  copy_data(H->strokes+4,cx,cy);
  free(data);
  free(points);
  return H;
}

static shape_t* make_trivial_strokes(int cx,int cy)
{
  int i;
  shape_t *H;

  H= calloc(1,sizeof(*H));
  for(i=0;i<9;i++)
    copy_data(H->strokes+i,cx,cy);
  free(data);
  free(points);
  return H;
}

shape_t *shape_thick_line()
{
  shape_t *H;
  stroke_t *T;
  H= calloc(1,sizeof(*H));
  points= malloc(4*sizeof(points[0]));

  T= H->strokes;
#define SS npoints=0
#define SE make_stroke(T); T++

                 // dy= -1
  SS; dpoint( 1,1,16); dpoint(1,0, 8); dpoint(0,1,8);  SE;
  SS; dpoint( 0,1,16);                   SE;
  SS; dpoint(-1,1,16); dpoint(0,1, 8); dpoint(-1,0,8);  SE;
                  // dy= 0
  SS; dpoint( 1, 0,16);                  SE;
  SS; dpoint( 0, 0,16);                  SE;
  SS; dpoint(-1, 0,16);                  SE;
                 // dy= 1
  SS; dpoint( 1,-1,16); dpoint(0,-1, 8);  SE;
  SS; dpoint( 0,-1,16);                  SE;
  SS; dpoint(-1,-1,16); dpoint(-1,0, 8);  SE;

  free(points);
  return H;
}

shape_t *shape_thin_line()
{
  shape_t *H;
  stroke_t *T;
  H= calloc(1,sizeof(*H));
  points= malloc(4*sizeof(points[0]));

  T= H->strokes;
                 // dy= -1
  npoints= 0; dpoint(-1,-1,16);  make_stroke(T); T++;
  npoints= 0; dpoint( 0,-1,16);  make_stroke(T); T++;
  npoints= 0; dpoint( 1,-1,16);  make_stroke(T); T++;
                  // dy= 0
  npoints= 0; dpoint(-1, 0,16);  make_stroke(T); T++;
  npoints= 0; dpoint( 0, 0,16);  make_stroke(T); T++;
  npoints= 0; dpoint( 1, 0,16);  make_stroke(T); T++;
                 // dy= 1
  npoints= 0; dpoint(-1, 1,16);  make_stroke(T); T++;
  npoints= 0; dpoint( 0, 1,16);  make_stroke(T); T++;
  npoints= 0; dpoint( 1, 1,16);  make_stroke(T); T++;
  
  free(points);
  return H;
}


shape_t *shape_new_circle(int r)
{
  alloc_data(2*r+1,2*r+1);
  memset(data,0,dw*dh);
  fillcircle(r);
  return make_all_strokes(r,r);
}

shape_t *shape_new_rectangle(int w,int h)
{
  alloc_data(w,h);
  memset(data,16,dw*dh);
  return make_all_strokes(w/2,h/2);
}

shape_t *shape_predef1()
{
  shape_t *S;
#if 1
  static const unsigned char dd[]=
   {  2,  2, 2, 
      2, 16, 2,
      2,  2, 2 };
  alloc_data(3,3);
#endif
#if 0
  static const unsigned char dd[]=
   {  2,  2, 2, 2, 
      2, 16, 8, 2,
      4,  4, 4, 4 };
  alloc_data(4,3);
#endif

  int i;
  for(i=0;i<sizeof(dd);i++) data[i]= dd[i];
  S= make_all_strokes(dw/2,dh/2);
#if 0
#ifdef GFX_X11
  int j;
  for(i=0;i<9;i++)
  {
    printf("stroke %d: ",i);
    for(j=0;j<S->strokes[i].len;j++)
      printf(" [%d %d %d] ", S->strokes[i].data[j].dx, S->strokes[i].data[j].dy, S->strokes[i].data[j].value);
    printf("\n");
  }
  printf("END\n");
#endif
#endif
  return S;
}

#ifdef SELF_TEST

int main(int argc,char **argv)
{
  mx= 1000; Mx= 0; My= 0; my= 1000;
  shape_new_circle(atoi(argv[1]));
  printf("minx= %d maxx= %d\n",mx,Mx);
  printf("miny= %d maxy= %d\n",my,My);
  prt_data();
}
#endif


void shape_free(shape_t *S)
{
  int i;
  for(i=0;i<9;i++)
   if (S->strokes[i].data) free(S->strokes[i].data);
  free(S);
}

shape_t *pen_new(int width)
{
  if (width<=2) 
//       return shape_new_circle(2);
     return shape_predef1();
  switch(width)
  {
  case 0: // return shape_new_rectangle(1,1); 
  case 1: return shape_predef1();
  case 2: return shape_new_circle(1);
  }
  return shape_new_circle(width);
}

