#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include "book.h"
#include "shufle.h"
#include "gfx.h"
#include "config.h"
#include "thumb.h"

static void *alloci(int n) { return malloc(sizeof(int)*n); }

#include "posmarker.ci"

static void permute(int N, int *_P, void *_A, size_t size)
{
  int *T,*P,j,i,prev,next;
  unsigned char *A,t;

  T= malloc(sizeof(int)*N*2);
  P= T+N;
  A= (unsigned char*) _A;
  for(i=0;i<N;i++)
  {
    P[i]= _P[i];
    T[P[i]]= i;
  }

  for(i=0;i<N;i++)
  {
    if (T[i]==i) continue;
    next= T[i];
    prev= P[i];

    for(j=0;j<size;j++)
    { t= A[i*size+j];
      A[i*size+j]= A[prev*size+j];
      A[prev*size+j]= t; }

    T[prev]= next;
    P[next]= prev;
  }
  free(T);
}


static void populate_page_list(book_t *fi)
{
  int i;
  unsigned int *td, tw, th;
  tw= config.thumb_w; th= config.thumb_h;
  td= alloci(tw*th);
  gfx_iconlist_updates(WpoLIST,0);

  for(i=0;i<fi->npages;i++)
  {
    char tbuf[20];
    gfx_iconlist_insert(WpoLIST,-1, " ", 
         posmarker[0], posmarker[1], posmarker+2, 0, NULL);
    if (make_page_thumbnail(fi,i,tw,th,td))
       make_default_thumbnail(tw,th,td);
    sprintf(tbuf,"%03d", i+1);
    gfx_iconlist_insert(WpoLIST,-1, tbuf, tw, th, td, 0, NULL);
  }
  gfx_iconlist_insert(WpoLIST,-1, "", 
       posmarker[0], posmarker[1], posmarker+2, 0, NULL);
  
  gfx_iconlist_updates(WpoLIST,1);
  free(td);
}

int reorder_pages(book_t *fi,int *Rn,int **Rp)
{
  int np,i,q, j,pos,it;
  int *fiperm, *perm, *sel, *S, nS, *pperm;
  if (!fi || fi->npages<2) return ;

  np= fi->npages;
  fiperm= alloci(np);
  for(i=0;i<np;i++) fiperm[i]= i;
  perm= alloci(np);
  sel= alloci(np);
  pperm= alloci(2*np+1);

  gfx_PAGEORDER();
  populate_page_list(fi);
  q= 0;
  while(!q)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case WpoOK: q=1; break;
    case WpoCANCEL: q= 2; break;
    case WpoLIST: 
      it= event.key;
      if (it<0 || it>=2*np+1) break;
      if (it%2==1) break;
      pos= it/2;
      for(i=0;i<np;i++) sel[i]= 0;
      gfx_iconlist_get_selection(WpoLIST, &nS, &S);
      for(i=0;i<nS;i++)
        if (S[i]%2==1) sel[S[i]/2]= 1;
      free(S);
      j= 0;
      for(i=0;i<pos;i++) if (!sel[i]) perm[j++]= i;
      for(i=0;i<np;i++) if (sel[i]) perm[j++]= i;
      for(i=pos;i<np;i++) if (!sel[i]) perm[j++]= i;
      for(i=0;i<np;i++) pperm[2*i+1]= perm[i]*2+1;
      for(i=0;i<=np;i++) pperm[2*i]= 2*i;
      gfx_iconlist_permute(WpoLIST, 2*np+1, pperm);
      permute(np, perm, fiperm, sizeof(int));
      gfx_iconlist_clear_selection(WpoLIST);
      break;
    }
  }
  gfx_leave_screen();
  free(perm);
  free(sel);
  free(pperm);
#ifdef GFX_X11
  if (q==1)
  {
    printf("Final Permutation: ");
    for(i=0;i<np;i++) printf("%d ", fiperm[i]);
    printf("\n");
  }
#endif
  if (q==1)
  {
    *Rn= np;
    *Rp= fiperm;
    return 0;
  }
  
  free(fiperm);
  return 1;
}
