#include <sys/types.h>
#include <dirent.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "pngio.h"
#include "resize_image.h"
#include "gfx.h"
#include "book.h"
#include "thumb.h"
#include "dbg.h"
#include "config.h"

void make_default_thumbnail(int w, int h, unsigned int *d)
{
  int i;
  for(i=0;i<w*h;i++) d[i]= 0x3355ee;
}

int make_page_thumbnail(book_t *fi, int pn, int w, int h, unsigned int *d)
{
  unsigned int mh, mw, *md; 
  unsigned int rw, rh, *rd;
  int i;
  int xo,yo,x,y;

  rw= w; rh= h; rd= d;

         // FIXME: this is the background
  for(i=0;i<w*h;i++) d[i]= 0xccCCff;

         // no thumbnail, don't try to make one here..
  if (book_read_image(fi, pn, 1, &rw, &rh, &rd, NULL)) return 1;

  if (rw>w || rh>h)
  {
    if (resize_image_argb(rd, rw, rh, w,h, &md, &mw, &mh))
      return 1;
    xo= (rw-mw)/2; yo= (rh-mh)/2;
    for(y=0;y<mh;y++)
      for(x=0;x<mw;x++)
       rd[(x+xo)+(y+yo)*rw]= md[x+y*mw];
    free(md);
  }
  return 0;
}


static int make_file_thumbnail
   (char *fn,int w,int h, unsigned int *d)
{
  book_t *B;
  int R;

  if (book_read(fn,&B)) { book_close(B); return 1; }
  R= make_page_thumbnail(B, 0, w, h, d);
  book_close(B);
  return R;
}

static int string_ends(char *fn, char *sfx)
{
  int Lf, Ls;
  Lf= strlen(fn);
  Ls= strlen(sfx);
  if (Ls>Lf) return 0;
  fn += Lf-Ls;
  return !strcmp(fn,sfx);
}


static int compar_file_ent(const void *_a, const void *_b)
{
  const char *const *a= _a;
  const char *const *b= _b;
  return strcmp(a[0],b[0]);
}

static int list_files(char **entries, int max,char *dn,char *skip)
{
  DIR *dir;
  int nentries;
  struct dirent *dent;
  char *fn;

  DPRINT(("Listing directory: %s\n", dn))
  if (chdir(dn)) return 0;
  dir= opendir(".");
  if (!dir) return 0; 
  DPRINT(("Files\n"))
  for(nentries=0;nentries<max;)
  {
    dent= readdir(dir);
    if (!dent) break;
    fn= dent->d_name;
    if (!strcmp(fn,".") || !strcmp(fn,"..")) continue;
    if (skip && !strcmp(fn,skip)) continue;
    if (!strcmp(fn,"about.not")) continue;
    if (!string_ends(fn,".not")) continue;
    entries[nentries++]= strdup(fn);
    DPRINT(("File: %s\n",fn))
  }
  closedir(dir);
  qsort(entries, nentries,sizeof(entries[0]),compar_file_ent);
  return nentries;
}

static char *new_file_name(char **files,int nfiles,char *buf)
{
  int i;
  int N,K;
  if (nfiles==0) { sprintf(buf,"001.not"); return buf; }
  K= atoi(files[0]);
  for(i=0;i<nfiles-1;i++)
  {
    N= K;
    K= atoi(files[i+1]);
    if (K-N>1) { sprintf(buf,"%03d.not", N+1); return buf; }
  }
  sprintf(buf,"%03d.not",nfiles+1); 
  return buf;
}

static char *make_new_file(char **files,int nfiles)
{
  char buf[200];
  new_file_name(files, nfiles,buf);
  create_note_file(buf);
  return strdup(buf);
}

static void populate_file_list(char **files, int nfiles)
{
  unsigned int iw, ih, *id;
  unsigned int *dd, *d;
  int i;
  char text_buf[200],*z;

  iw= config.thumb_w; ih= config.thumb_h;
  id= malloc(sizeof(*id)*iw*ih);
  dd= NULL;

  gfx_iconlist_updates(WftLIST,0);
  for(i=0;i<nfiles;i++)
  {
    strncpy(text_buf, files[i], sizeof(text_buf));
    text_buf[sizeof(text_buf)-1]=0;
    z= strchr(text_buf,'.');
    if (z) *z= 0;

    if (make_file_thumbnail(files[i], iw, ih, id)) 
    {
       if (!dd)
       {
         dd= malloc(sizeof(*dd)*iw*ih);
         make_default_thumbnail(iw,ih, dd);
       }
       d= dd;
    }
    else
    {
       d= id;
    }
    gfx_iconlist_insert(WftLIST,-1,text_buf,iw,ih,d, 0, files[i]);
  }
  gfx_iconlist_updates(WftLIST,1);
  if (dd) free(dd);
  free(id);
}

char* choose_file
   (int tw,int th, char *dn,char *cfn)
{
  char *RV;
  char* files[100]; int nfiles;
  gfx_event_t event;
  int i;

  gfx_FILETHUMBS();
  nfiles= list_files(files, 100, dn, NULL);
  populate_file_list(files, nfiles);

  RV= NULL;
again:
  if (!gfx_event(NULL,&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT:
  case WftCLOSE: break;
  case WftNEW: if (!(RV= make_new_file(files,nfiles))) goto again; break;
  case WftLIST: RV= event.usr; break;
  default: goto again;
  }
  gfx_leave_screen();
  for(i=0;i<nfiles;i++)
    if (files[i]!=RV) free(files[i]);
  return RV;
}


void del_all(char *dst)
{
#ifdef GFX_ANDROID
  char fn[2000];
  struct dirent *de;
  DIR *dir;
  dir= opendir(dst);
  if (!dir) return ;
  while((de=readdir(dir)))
  {
    if (!strcmp(de->d_name,".") || !strcmp(de->d_name,"..")) continue;
    sprintf(fn,"%s/%s",dst, de->d_name);
    del_all(fn);
    rmdir(fn);
    unlink(fn);
  }
  closedir(dir);
#endif
}

